@@ -30,41 +30,41 @@
 #include <memory>
 #include <string>
 
-#include "builtin_argparse.h"
-#include "builtin_bg.h"
-#include "builtin_bind.h"
-#include "builtin_block.h"
-#include "builtin_builtin.h"
-#include "builtin_cd.h"
-#include "builtin_command.h"
-#include "builtin_commandline.h"
-#include "builtin_complete.h"
-#include "builtin_contains.h"
-#include "builtin_disown.h"
-#include "builtin_echo.h"
-#include "builtin_emit.h"
-#include "builtin_eval.h"
-#include "builtin_exit.h"
-#include "builtin_fg.h"
-#include "builtin_functions.h"
-#include "builtin_history.h"
-#include "builtin_jobs.h"
-#include "builtin_math.h"
-#include "builtin_printf.h"
-#include "builtin_pwd.h"
-#include "builtin_random.h"
-#include "builtin_read.h"
-#include "builtin_realpath.h"
-#include "builtin_return.h"
-#include "builtin_set.h"
-#include "builtin_set_color.h"
-#include "builtin_source.h"
-#include "builtin_status.h"
-#include "builtin_string.h"
-#include "builtin_test.h"
-#include "builtin_type.h"
-#include "builtin_ulimit.h"
-#include "builtin_wait.h"
+#include "builtins/argparse.h"
+#include "builtins/bg.h"
+#include "builtins/bind.h"
+#include "builtins/block.h"
+#include "builtins/builtin.h"
+#include "builtins/cd.h"
+#include "builtins/command.h"
+#include "builtins/commandline.h"
+#include "builtins/complete.h"
+#include "builtins/contains.h"
+#include "builtins/disown.h"
+#include "builtins/echo.h"
+#include "builtins/emit.h"
+#include "builtins/eval.h"
+#include "builtins/exit.h"
+#include "builtins/fg.h"
+#include "builtins/functions.h"
+#include "builtins/history.h"
+#include "builtins/jobs.h"
+#include "builtins/math.h"
+#include "builtins/printf.h"
+#include "builtins/pwd.h"
+#include "builtins/random.h"
+#include "builtins/read.h"
+#include "builtins/realpath.h"
+#include "builtins/return.h"
+#include "builtins/set.h"
+#include "builtins/set_color.h"
+#include "builtins/source.h"
+#include "builtins/status.h"
+#include "builtins/string.h"
+#include "builtins/test.h"
+#include "builtins/type.h"
+#include "builtins/ulimit.h"
+#include "builtins/wait.h"
 #include "common.h"
 #include "complete.h"
 #include "exec.h"
@@ -80,14 +80,6 @@
 #include "wgetopt.h"
 #include "wutil.h"  // IWYU pragma: keep
 
-bool builtin_data_t::operator<(const wcstring &other) const {
-    return std::wcscmp(this->name, other.c_str()) < 0;
-}
-
-bool builtin_data_t::operator<(const builtin_data_t *other) const {
-    return std::wcscmp(this->name, other->name) < 0;
-}
-
 /// Counts the number of arguments in the specified null-terminated array
 int builtin_count_args(const wchar_t *const *argv) {
     int argc;
@@ -101,10 +93,10 @@ int builtin_count_args(const wchar_t *const *argv) {
 
 /// This function works like wperror, but it prints its result into the streams.err string instead
 /// to stderr. Used by the builtin commands.
-void builtin_wperror(const wchar_t *s, io_streams_t &streams) {
+void builtin_wperror(const wchar_t *program_name, io_streams_t &streams) {
     char *err = std::strerror(errno);
-    if (s != nullptr) {
-        streams.err.append(s);
+    if (program_name != nullptr) {
+        streams.err.append(program_name);
         streams.err.append(L": ");
     }
     if (err != nullptr) {
@@ -119,8 +111,8 @@ static const struct woption long_options[] = {{L"help", no_argument, nullptr, 'h
                                               {nullptr, 0, nullptr, 0}};
 
 int parse_help_only_cmd_opts(struct help_only_cmd_opts_t &opts, int *optind, int argc,
-                             wchar_t **argv, parser_t &parser, io_streams_t &streams) {
-    wchar_t *cmd = argv[0];
+                             const wchar_t **argv, parser_t &parser, io_streams_t &streams) {
+    const wchar_t *cmd = argv[0];
     int opt;
     wgetopter_t w;
     while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, nullptr)) != -1) {
@@ -166,25 +158,31 @@ void builtin_print_help(parser_t &parser, const io_streams_t &streams, const wch
         ios.push_back(std::make_shared<io_fd_t>(STDOUT_FILENO, STDERR_FILENO));
     }
     auto res = parser.eval(cmd, ios);
-    if (res.status.exit_code() == 2) {
-        streams.err.append_format(BUILTIN_ERR_MISSING_HELP, name_esc.c_str());
+    if (res.status.normal_exited() && res.status.exit_code() == 2) {
+        streams.err.append_format(BUILTIN_ERR_MISSING_HELP, name_esc.c_str(), name_esc.c_str());
     }
 }
 
 /// Perform error reporting for encounter with unknown option.
 void builtin_unknown_option(parser_t &parser, io_streams_t &streams, const wchar_t *cmd,
-                            const wchar_t *opt) {
+                            const wchar_t *opt, bool print_hints) {
     streams.err.append_format(BUILTIN_ERR_UNKNOWN, cmd, opt);
-    builtin_print_error_trailer(parser, streams.err, cmd);
+    if (print_hints) {
+        builtin_print_error_trailer(parser, streams.err, cmd);
+    }
 }
 
 /// Perform error reporting for encounter with missing argument.
 void builtin_missing_argument(parser_t &parser, io_streams_t &streams, const wchar_t *cmd,
                               const wchar_t *opt, bool print_hints) {
     if (opt[0] == L'-' && opt[1] != L'-') {
+        // if c in -qc '-qc' is missing the argument, now opt is just 'c'
         opt += std::wcslen(opt) - 1;
-    }
-    streams.err.append_format(BUILTIN_ERR_MISSING, cmd, opt);
+        // now prepend - to output -c
+        streams.err.append_format(BUILTIN_ERR_MISSING, cmd, wcstring(L"-").append(opt).c_str());
+    } else
+        streams.err.append_format(BUILTIN_ERR_MISSING, cmd, opt);
+
     if (print_hints) {
         builtin_print_error_trailer(parser, streams.err, cmd);
     }
@@ -204,7 +202,7 @@ void builtin_print_error_trailer(parser_t &parser, output_stream_t &b, const wch
 
 /// A generic bultin that only supports showing a help message. This is only a placeholder that
 /// prints the help message. Useful for commands that live in the parser.
-static maybe_t<int> builtin_generic(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
+static maybe_t<int> builtin_generic(parser_t &parser, io_streams_t &streams, const wchar_t **argv) {
     const wchar_t *cmd = argv[0];
     int argc = builtin_count_args(argv);
     help_only_cmd_opts_t opts;
@@ -231,7 +229,7 @@ static maybe_t<int> builtin_generic(parser_t &parser, io_streams_t &streams, wch
 // Since this is just for counting, it can be massive.
 #define COUNT_CHUNK_SIZE (512 * 256)
 /// Implementation of the builtin count command, used to count the number of arguments sent to it.
-static maybe_t<int> builtin_count(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
+static maybe_t<int> builtin_count(parser_t &parser, io_streams_t &streams, const wchar_t **argv) {
     UNUSED(parser);
     int argc = 0;
 
@@ -249,7 +247,7 @@ static maybe_t<int> builtin_count(parser_t &parser, io_streams_t &streams, wchar
                 return STATUS_CMD_ERROR;
             }
             for (int i = 0; i < n; i++) {
-                if (buf[i] == L'\n') {
+                if (buf[i] == '\n') {
                     argc++;
                 }
             }
@@ -266,7 +264,7 @@ static maybe_t<int> builtin_count(parser_t &parser, io_streams_t &streams, wchar
 /// This function handles both the 'continue' and the 'break' builtins that are used for loop
 /// control.
 static maybe_t<int> builtin_break_continue(parser_t &parser, io_streams_t &streams,
-                                           wchar_t **argv) {
+                                           const wchar_t **argv) {
     int is_break = (std::wcscmp(argv[0], L"break") == 0);
     int argc = builtin_count_args(argv);
 
@@ -277,6 +275,7 @@ static maybe_t<int> builtin_break_continue(parser_t &parser, io_streams_t &strea
     }
 
     // Paranoia: ensure we have a real loop.
+    // This is checked in the AST but we may be invoked dynamically, e.g. just via "eval break".
     bool has_loop = false;
     for (const auto &b : parser.blocks()) {
         if (b.type() == block_type_t::while_block || b.type() == block_type_t::for_block) {
@@ -297,8 +296,9 @@ static maybe_t<int> builtin_break_continue(parser_t &parser, io_streams_t &strea
 }
 
 /// Implementation of the builtin breakpoint command, used to launch the interactive debugger.
-static maybe_t<int> builtin_breakpoint(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
-    wchar_t *cmd = argv[0];
+static maybe_t<int> builtin_breakpoint(parser_t &parser, io_streams_t &streams,
+                                       const wchar_t **argv) {
+    const wchar_t *cmd = argv[0];
     if (argv[1] != nullptr) {
         streams.err.append_format(BUILTIN_ERR_ARG_COUNT1, cmd, 0, builtin_count_args(argv) - 1);
         return STATUS_INVALID_ARGS;
@@ -323,21 +323,21 @@ static maybe_t<int> builtin_breakpoint(parser_t &parser, io_streams_t &streams,
     return parser.get_last_status();
 }
 
-maybe_t<int> builtin_true(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
+static maybe_t<int> builtin_true(parser_t &parser, io_streams_t &streams, const wchar_t **argv) {
     UNUSED(parser);
     UNUSED(streams);
     UNUSED(argv);
     return STATUS_CMD_OK;
 }
 
-maybe_t<int> builtin_false(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
+static maybe_t<int> builtin_false(parser_t &parser, io_streams_t &streams, const wchar_t **argv) {
     UNUSED(parser);
     UNUSED(streams);
     UNUSED(argv);
     return STATUS_CMD_ERROR;
 }
 
-maybe_t<int> builtin_gettext(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
+static maybe_t<int> builtin_gettext(parser_t &parser, io_streams_t &streams, const wchar_t **argv) {
     UNUSED(parser);
     UNUSED(streams);
     for (int i = 1; i < builtin_count_args(argv); i++) {
@@ -353,29 +353,27 @@ maybe_t<int> builtin_gettext(parser_t &parser, io_streams_t &streams, wchar_t **
 // Data about all the builtin commands in fish.
 // Functions that are bound to builtin_generic are handled directly by the parser.
 // NOTE: These must be kept in sorted order!
-static const builtin_data_t builtin_datas[] = {
+static constexpr builtin_data_t builtin_datas[] = {
     {L".", &builtin_source, N_(L"Evaluate contents of file")},
     {L":", &builtin_true, N_(L"Return a successful result")},
     {L"[", &builtin_test, N_(L"Test a condition")},
     {L"_", &builtin_gettext, N_(L"Translate a string")},
-    {L"and", &builtin_generic, N_(L"Execute command if previous command succeeded")},
+    {L"and", &builtin_generic, N_(L"Run command if last command succeeded")},
     {L"argparse", &builtin_argparse, N_(L"Parse options in fish script")},
     {L"begin", &builtin_generic, N_(L"Create a block of code")},
     {L"bg", &builtin_bg, N_(L"Send job to background")},
     {L"bind", &builtin_bind, N_(L"Handle fish key bindings")},
     {L"block", &builtin_block, N_(L"Temporarily block delivery of events")},
     {L"break", &builtin_break_continue, N_(L"Stop the innermost loop")},
-    {L"breakpoint", &builtin_breakpoint,
-     N_(L"Temporarily halt execution of a script and launch an interactive debug prompt")},
-    {L"builtin", &builtin_builtin, N_(L"Run a builtin command instead of a function")},
-    {L"case", &builtin_generic, N_(L"Conditionally execute a block of commands")},
+    {L"breakpoint", &builtin_breakpoint, N_(L"Halt execution and start debug prompt")},
+    {L"builtin", &builtin_builtin, N_(L"Run a builtin specifically")},
+    {L"case", &builtin_generic, N_(L"Block of code to run conditionally")},
     {L"cd", &builtin_cd, N_(L"Change working directory")},
-    {L"command", &builtin_command, N_(L"Run a program instead of a function or builtin")},
+    {L"command", &builtin_command, N_(L"Run a command specifically")},
     {L"commandline", &builtin_commandline, N_(L"Set or get the commandline")},
     {L"complete", &builtin_complete, N_(L"Edit command specific completions")},
     {L"contains", &builtin_contains, N_(L"Search for a specified string in a list")},
-    {L"continue", &builtin_break_continue,
-     N_(L"Skip the rest of the current lap of the innermost loop")},
+    {L"continue", &builtin_break_continue, N_(L"Skip over remaining innermost loop")},
     {L"count", &builtin_count, N_(L"Count the number of arguments")},
     {L"disown", &builtin_disown, N_(L"Remove job from job list")},
     {L"echo", &builtin_echo, N_(L"Print arguments")},
@@ -400,22 +398,23 @@ static const builtin_data_t builtin_datas[] = {
     {L"pwd", &builtin_pwd, N_(L"Print the working directory")},
     {L"random", &builtin_random, N_(L"Generate random number")},
     {L"read", &builtin_read, N_(L"Read a line of input into variables")},
-    {L"realpath", &builtin_realpath, N_(L"Convert path to absolute path without symlinks")},
+    {L"realpath", &builtin_realpath, N_(L"Show absolute path sans symlinks")},
     {L"return", &builtin_return, N_(L"Stop the currently evaluated function")},
     {L"set", &builtin_set, N_(L"Handle environment variables")},
     {L"set_color", &builtin_set_color, N_(L"Set the terminal color")},
     {L"source", &builtin_source, N_(L"Evaluate contents of file")},
     {L"status", &builtin_status, N_(L"Return status information about fish")},
     {L"string", &builtin_string, N_(L"Manipulate strings")},
-    {L"switch", &builtin_generic, N_(L"Conditionally execute a block of commands")},
+    {L"switch", &builtin_generic, N_(L"Conditionally run blocks of code")},
     {L"test", &builtin_test, N_(L"Test a condition")},
     {L"time", &builtin_generic, N_(L"Measure how long a command or block takes")},
     {L"true", &builtin_true, N_(L"Return a successful result")},
     {L"type", &builtin_type, N_(L"Check if a thing is a thing")},
-    {L"ulimit", &builtin_ulimit, N_(L"Set or get the shells resource usage limits")},
+    {L"ulimit", &builtin_ulimit, N_(L"Get/set resource usage limits")},
     {L"wait", &builtin_wait, N_(L"Wait for background processes completed")},
     {L"while", &builtin_generic, N_(L"Perform a command multiple times")},
 };
+ASSERT_SORTED_BY_NAME(builtin_datas);
 
 #define BUILTIN_COUNT (sizeof builtin_datas / sizeof *builtin_datas)
 
@@ -428,21 +427,13 @@ static const builtin_data_t builtin_datas[] = {
 ///    Pointer to a builtin_data_t
 ///
 static const builtin_data_t *builtin_lookup(const wcstring &name) {
-    const builtin_data_t *array_end = builtin_datas + BUILTIN_COUNT;
-    const builtin_data_t *found = std::lower_bound(builtin_datas, array_end, name);
-    if (found != array_end && name == found->name) {
-        return found;
-    }
-    return nullptr;
+    return get_by_sorted_name(name.c_str(), builtin_datas);
 }
 
 /// Initialize builtin data.
 void builtin_init() {
-    for (size_t i = 0; i < BUILTIN_COUNT; i++) {
-        const wchar_t *name = builtin_datas[i].name;
-        intern_static(name);
-        assert((i == 0 || std::wcscmp(builtin_datas[i - 1].name, name) < 0) &&
-               "builtins are not sorted alphabetically");
+    for (const auto &builtin_data : builtin_datas) {
+        intern_static(builtin_data.name);
     }
 }
 
@@ -452,40 +443,49 @@ bool builtin_exists(const wcstring &cmd) { return static_cast<bool>(builtin_look
 /// Is the command a keyword we need to special-case the handling of `-h` and `--help`.
 static const wchar_t *const help_builtins[] = {L"for", L"while",  L"function", L"if",
                                                L"end", L"switch", L"case"};
-static bool cmd_needs_help(const wchar_t *cmd) { return contains(help_builtins, cmd); }
+static bool cmd_needs_help(const wcstring &cmd) { return contains(help_builtins, cmd); }
 
 /// Execute a builtin command
-proc_status_t builtin_run(parser_t &parser, wchar_t **argv, io_streams_t &streams) {
-    UNUSED(parser);
-    UNUSED(streams);
-    if (argv == nullptr || argv[0] == nullptr)
-        return proc_status_t::from_exit_code(STATUS_INVALID_ARGS);
+proc_status_t builtin_run(parser_t &parser, const wcstring_list_t &argv, io_streams_t &streams) {
+    if (argv.empty()) return proc_status_t::from_exit_code(STATUS_INVALID_ARGS);
+    const wcstring &cmdname = argv.front();
 
     // We can be handed a keyword by the parser as if it was a command. This happens when the user
     // follows the keyword by `-h` or `--help`. Since it isn't really a builtin command we need to
     // handle displaying help for it here.
-    if (argv[1] && !argv[2] && parse_util_argument_is_help(argv[1]) && cmd_needs_help(argv[0])) {
-        builtin_print_help(parser, streams, argv[0]);
+    if (argv.size() == 2 && parse_util_argument_is_help(argv[1]) && cmd_needs_help(cmdname)) {
+        builtin_print_help(parser, streams, cmdname.c_str());
         return proc_status_t::from_exit_code(STATUS_CMD_OK);
     }
 
-    if (const builtin_data_t *data = builtin_lookup(argv[0])) {
-        maybe_t<int> ret = data->func(parser, streams, argv);
-        if (!ret) {
-            return proc_status_t::empty();
-        }
+    if (const builtin_data_t *data = builtin_lookup(cmdname)) {
+        // Construct the permutable argv array which the builtin expects, and execute the builtin.
+        null_terminated_array_t<wchar_t> argv_arr(argv);
+        maybe_t<int> builtin_ret = data->func(parser, streams, argv_arr.get());
+
+        // Flush our out and error streams, and check for their errors.
+        int out_ret = streams.out.flush_and_check_error();
+        int err_ret = streams.err.flush_and_check_error();
+
+        // Resolve our status code.
+        // If the builtin itself produced an error, use that error.
+        // Otherwise use any errors from writing to out and writing to err, in that order.
+        int code = builtin_ret ? *builtin_ret : 0;
+        if (code == 0) code = out_ret;
+        if (code == 0) code = err_ret;
 
         // The exit code is cast to an 8-bit unsigned integer, so saturate to 255. Otherwise,
         // multiples of 256 are reported as 0.
-        int code = ret.value();
-        if (code > 255) {
-            code = 255;
-        }
+        if (code > 255) code = 255;
 
+        // Handle the case of an empty status.
+        if (code == 0 && !builtin_ret.has_value()) {
+            return proc_status_t::empty();
+        }
         return proc_status_t::from_exit_code(code);
     }
 
-    FLOGF(error, UNKNOWN_BUILTIN_ERR_MSG, argv[0]);
+    FLOGF(error, UNKNOWN_BUILTIN_ERR_MSG, cmdname.c_str());
     return proc_status_t::from_exit_code(STATUS_CMD_ERROR);
 }
 
