@@ -56,9 +56,10 @@
 #include "wutil.h"  // IWYU pragma: keep
 
 // Keep after "common.h"
-#ifdef __BSD__
+#ifdef HAVE_SYS_SYSCTL_H
 #include <sys/sysctl.h>
-#elif defined(__APPLE__)
+#endif
+#if defined(__APPLE__)
 #include <mach-o/dyld.h>
 #endif
 
@@ -87,8 +88,8 @@ static relaxed_atomic_t<wchar_t> obfuscation_read_char;
 wchar_t get_obfuscation_read_char() { return obfuscation_read_char; }
 
 bool g_profiling_active = false;
+
 const wchar_t *program_name;
-std::atomic<int> debug_level{1};  // default maximum debug output level (errors and warnings)
 
 /// Be able to restore the term's foreground process group.
 /// This is set during startup and not modified after.
@@ -230,7 +231,7 @@ bool is_windows_subsystem_for_linux() {
     debug_shared(msg_level, L"Backtrace:\n" + join_strings(bt, L'\n') + L'\n');
 }
 
-#else  // HAVE_BACKTRACE_SYMBOLS
+#else   // HAVE_BACKTRACE_SYMBOLS
 
 [[gnu::noinline]] void show_stackframe(const wchar_t msg_level, int, int) {
     debug_shared(msg_level, L"Sorry, but your system does not support backtraces");
@@ -242,7 +243,7 @@ bool is_windows_subsystem_for_linux() {
 /// alignment must be a power of 2 and in range [1, 64].
 /// This is intended to return the end point of the "unaligned prefix" of a vectorized loop.
 template <size_t Align>
-inline const char *align_start(const char *start, size_t len) {
+static inline const char *align_start(const char *start, size_t len) {
     static_assert(Align >= 1 && Align <= 64, "Alignment must be in range [1, 64]");
     static_assert((Align & (Align - 1)) == 0, "Alignment must be power of 2");
     uintptr_t startu = reinterpret_cast<uintptr_t>(start);
@@ -262,7 +263,7 @@ inline const char *align_start(const char *start, size_t len) {
 /// If there is no such pointer, return \p start.
 /// This is intended to be the start point of the "unaligned suffix" of a vectorized loop.
 template <size_t Align>
-inline const char *align_end(const char *start, size_t len) {
+static inline const char *align_end(const char *start, size_t len) {
     static_assert(Align >= 1 && Align <= 64, "Alignment must be in range [1, 64]");
     static_assert((Align & (Align - 1)) == 0, "Alignment must be power of 2");
     // How much do we have to subtract to align it? Its value, mod Align.
@@ -501,9 +502,7 @@ void append_format(wcstring &str, const wchar_t *format, ...) {
     va_end(va);
 }
 
-wchar_t *quote_end(const wchar_t *pos) {
-    wchar_t c = *pos;
-
+const wchar_t *quote_end(const wchar_t *pos, wchar_t quote) {
     while (true) {
         pos++;
 
@@ -513,14 +512,24 @@ wchar_t *quote_end(const wchar_t *pos) {
             pos++;
             if (!*pos) return nullptr;
         } else {
-            if (*pos == c) {
-                return const_cast<wchar_t *>(pos);
+            if (*pos == quote ||
+                // Command substitutions also end a double quoted string.  This is how we
+                // support command substitutions inside double quotes.
+                (quote == L'"' && *pos == L'$' && *(pos + 1) == L'(')) {
+                return pos;
             }
         }
     }
     return nullptr;
 }
 
+const wchar_t *comment_end(const wchar_t *pos) {
+    do {
+        pos++;
+    } while (*pos && *pos != L'\n');
+    return pos;
+}
+
 void fish_setlocale() {
     // Use various Unicode symbols if they can be encoded using the current locale, else a simple
     // ASCII char alternative. All of the can_be_encoded() invocations should return the same
@@ -547,7 +556,7 @@ void fish_setlocale() {
         obfuscation_read_char = L'*';
     } else {
         if (can_be_encoded(L'\u23CE')) {
-            omitted_newline_str = L"\u23CE";
+            omitted_newline_str = L"\u23CE";  // "return symbol" (⏎)
             omitted_newline_width = 1;
         } else {
             omitted_newline_str = L"^J";
@@ -600,57 +609,14 @@ bool should_suppress_stderr_for_tests() {
 static void debug_shared(const wchar_t level, const wcstring &msg) {
     pid_t current_pid;
     if (!is_forked_child()) {
-        std::fwprintf(stderr, L"<%lc> %ls: %ls\n", static_cast<unsigned long>(level), program_name,
-                      msg.c_str());
+        std::fwprintf(stderr, L"<%lc> %ls: %ls\n", level, program_name, msg.c_str());
     } else {
         current_pid = getpid();
-        std::fwprintf(stderr, L"<%lc> %ls: %d: %ls\n", static_cast<unsigned long>(level),
-                      program_name, current_pid, msg.c_str());
+        std::fwprintf(stderr, L"<%lc> %ls: %d: %ls\n", level, program_name, current_pid,
+                      msg.c_str());
     }
 }
 
-void debug_safe(int level, const char *msg, const char *param1, const char *param2,
-                const char *param3, const char *param4, const char *param5, const char *param6,
-                const char *param7, const char *param8, const char *param9, const char *param10,
-                const char *param11, const char *param12) {
-    const char *const params[] = {param1, param2, param3, param4,  param5,  param6,
-                                  param7, param8, param9, param10, param11, param12};
-    if (!msg) return;
-
-    // Can't call fwprintf, that may allocate memory Just call write() over and over.
-    if (level > debug_level) return;
-    int errno_old = errno;
-
-    size_t param_idx = 0;
-    const char *cursor = msg;
-    while (*cursor != '\0') {
-        const char *end = std::strchr(cursor, '%');
-        if (end == nullptr) end = cursor + std::strlen(cursor);
-
-        ignore_result(write(STDERR_FILENO, cursor, end - cursor));
-
-        if (end[0] == '%' && end[1] == 's') {
-            // Handle a format string.
-            assert(param_idx < sizeof params / sizeof *params);
-            const char *format = params[param_idx++];
-            if (!format) format = "(null)";
-            ignore_result(write(STDERR_FILENO, format, std::strlen(format)));
-            cursor = end + 2;
-        } else if (end[0] == '\0') {
-            // Must be at the end of the string.
-            cursor = end;
-        } else {
-            // Some other format specifier, just skip it.
-            cursor = end + 1;
-        }
-    }
-
-    // We always append a newline.
-    ignore_result(write(STDERR_FILENO, "\n", 1));
-
-    errno = errno_old;
-}
-
 // Careful to not negate LLONG_MIN.
 static unsigned long long absolute_value(long long x) {
     if (x >= 0) return static_cast<unsigned long long>(x);
@@ -659,7 +625,7 @@ static unsigned long long absolute_value(long long x) {
 }
 
 template <typename CharT>
-void format_safe_impl(CharT *buff, size_t size, unsigned long long val) {
+static void format_safe_impl(CharT *buff, size_t size, unsigned long long val) {
     size_t idx = 0;
     if (val == 0) {
         buff[idx++] = '0';
@@ -897,6 +863,22 @@ static bool unescape_string_var(const wchar_t *in, wcstring *out) {
     return true;
 }
 
+wcstring escape_string_for_double_quotes(wcstring in) {
+    // We need to escape backslashes, double quotes, and dollars only.
+    wcstring result = std::move(in);
+    size_t idx = result.size();
+    while (idx--) {
+        switch (result[idx]) {
+            case L'\\':
+            case L'$':
+            case L'"':
+                result.insert(idx, 1, L'\\');
+                break;
+        }
+    }
+    return result;
+}
+
 /// Escape a string in a fashion suitable for using in fish script. Store the result in out_str.
 static void escape_string_script(const wchar_t *orig_in, size_t in_len, wcstring &out,
                                  escape_flags_t flags) {
@@ -1156,7 +1138,7 @@ static maybe_t<wchar_t> string_last_char(const wcstring &str) {
 }
 
 /// Given a null terminated string starting with a backslash, read the escape as if it is unquoted,
-/// appending to result. Return the number of characters consumed, or 0 on error.
+/// appending to result. Return the number of characters consumed, or none on error.
 maybe_t<size_t> read_unquoted_escape(const wchar_t *input, wcstring *result, bool allow_incomplete,
                                      bool unescape_special) {
     assert(input[0] == L'\\' && "Not an escape");
@@ -1237,6 +1219,8 @@ maybe_t<size_t> read_unquoted_escape(const wchar_t *input, wcstring *result, boo
             for (size_t i = 0; i < chars; i++) {
                 long d = convert_digit(input[in_pos], base);
                 if (d < 0) {
+                    // If we have no digit, this is a tokenizer error.
+                    if (i == 0) errored = true;
                     break;
                 }
 
@@ -1244,7 +1228,7 @@ maybe_t<size_t> read_unquoted_escape(const wchar_t *input, wcstring *result, boo
                 in_pos++;
             }
 
-            if (res <= max_val) {
+            if (!errored && res <= max_val) {
                 result_char_or_none =
                     static_cast<wchar_t>((byte_literal ? ENCODE_DIRECT_BASE : 0) + res);
             } else {
@@ -1326,7 +1310,7 @@ maybe_t<size_t> read_unquoted_escape(const wchar_t *input, wcstring *result, boo
 }
 
 /// Returns the unescaped version of input_str into output_str (by reference). Returns true if
-/// successful. If false, the contents of output_str are undefined (!).
+/// successful. If false, the contents of output_str are unchanged.
 static bool unescape_string_internal(const wchar_t *const input, const size_t input_len,
                                      wcstring *output_str, unescape_flags_t flags) {
     // Set up result string, which we'll swap with the output on success.
@@ -1416,8 +1400,12 @@ static bool unescape_string_internal(const wchar_t *const input, const size_t in
                 }
                 case L'$': {
                     if (unescape_special) {
-                        to_append_or_none = VARIABLE_EXPAND;
-                        vars_or_seps.push_back(input_position);
+                        bool is_cmdsub =
+                            input_position + 1 < input_len && input[input_position + 1] == L'(';
+                        if (!is_cmdsub) {
+                            to_append_or_none = VARIABLE_EXPAND;
+                            vars_or_seps.push_back(input_position);
+                        }
                     }
                     break;
                 }
@@ -1600,12 +1588,12 @@ bool unescape_string_in_place(wcstring *str, unescape_flags_t escape_special) {
     return success;
 }
 
-bool unescape_string(const wchar_t *input, wcstring *output, unescape_flags_t escape_special,
-                     escape_string_style_t style) {
+bool unescape_string(const wchar_t *input, size_t len, wcstring *output,
+                     unescape_flags_t escape_special, escape_string_style_t style) {
     bool success = false;
     switch (style) {
         case STRING_STYLE_SCRIPT: {
-            success = unescape_string_internal(input, std::wcslen(input), output, escape_special);
+            success = unescape_string_internal(input, len, output, escape_special);
             break;
         }
         case STRING_STYLE_URL: {
@@ -1626,30 +1614,14 @@ bool unescape_string(const wchar_t *input, wcstring *output, unescape_flags_t es
     return success;
 }
 
+bool unescape_string(const wchar_t *input, wcstring *output, unescape_flags_t escape_special,
+                     escape_string_style_t style) {
+    return unescape_string(input, std::wcslen(input), output, escape_special, style);
+}
+
 bool unescape_string(const wcstring &input, wcstring *output, unescape_flags_t escape_special,
                      escape_string_style_t style) {
-    bool success = false;
-    switch (style) {
-        case STRING_STYLE_SCRIPT: {
-            success = unescape_string_internal(input.c_str(), input.size(), output, escape_special);
-            break;
-        }
-        case STRING_STYLE_URL: {
-            success = unescape_string_url(input.c_str(), output);
-            break;
-        }
-        case STRING_STYLE_VAR: {
-            success = unescape_string_var(input.c_str(), output);
-            break;
-        }
-        case STRING_STYLE_REGEX: {
-            // unescaping PCRE2 is not needed/supported, the PCRE2 engine is responsible for that
-            success = false;
-            break;
-        }
-    }
-    if (!success) output->clear();
-    return success;
+    return unescape_string(input.c_str(), input.size(), output, escape_special, style);
 }
 
 [[gnu::noinline]] void bugreport() {
@@ -1674,7 +1646,7 @@ wcstring format_size(long long sz) {
             if (sz < (1024 * 1024) || !sz_name[i + 1]) {
                 long isz = (static_cast<long>(sz)) / 1024;
                 if (isz > 9)
-                    result.append(format_string(L"%d%ls", isz, sz_name[i]));
+                    result.append(format_string(L"%ld%ls", isz, sz_name[i]));
                 else
                     result.append(
                         format_string(L"%.1f%ls", static_cast<double>(sz) / 1024, sz_name[i]));
@@ -1745,13 +1717,10 @@ void format_size_safe(char buff[128], unsigned long long sz) {
     }
 }
 
-/// Return the number of seconds from the UNIX epoch, with subsecond precision. This function uses
-/// the gettimeofday function and will have the same precision as that function.
 double timef() {
     struct timeval tv;
     assert_with_errno(gettimeofday(&tv, nullptr) != -1);
-    // return (double)tv.tv_sec + 0.000001 * tv.tv_usec;
-    return static_cast<double>(tv.tv_sec) + 1e-6 * tv.tv_usec;
+    return static_cast<timepoint_t>(tv.tv_sec) + 1e-6 * tv.tv_usec;
 }
 
 void exit_without_destructors(int code) { _exit(code); }
@@ -1805,7 +1774,7 @@ void restore_term_foreground_process_group_for_exit() {
 bool is_main_thread() { return thread_id() == 1; }
 
 void assert_is_main_thread(const char *who) {
-    if (!is_main_thread() && !thread_asserts_cfg_for_testing) {
+    if (!likely(is_main_thread()) && !unlikely(thread_asserts_cfg_for_testing)) {
         FLOGF(error, L"%s called off of main thread.", who);
         FLOGF(error, L"Break on debug_thread_error to debug.");
         debug_thread_error();
@@ -1813,7 +1782,7 @@ void assert_is_main_thread(const char *who) {
 }
 
 void assert_is_not_forked_child(const char *who) {
-    if (is_forked_child()) {
+    if (unlikely(is_forked_child())) {
         FLOGF(error, L"%s called in a forked child.", who);
         FLOG(error, L"Break on debug_thread_error to debug.");
         debug_thread_error();
@@ -1821,7 +1790,7 @@ void assert_is_not_forked_child(const char *who) {
 }
 
 void assert_is_background_thread(const char *who) {
-    if (is_main_thread() && !thread_asserts_cfg_for_testing) {
+    if (unlikely(is_main_thread()) && !unlikely(thread_asserts_cfg_for_testing)) {
         FLOGF(error, L"%s called on the main thread (may block!).", who);
         FLOG(error, L"Break on debug_thread_error to debug.");
         debug_thread_error();
@@ -1831,7 +1800,7 @@ void assert_is_background_thread(const char *who) {
 void assert_is_locked(std::mutex &mutex, const char *who, const char *caller) {
     // Note that std::mutex.try_lock() is allowed to return false when the mutex isn't
     // actually locked; fortunately we are checking the opposite so we're safe.
-    if (mutex.try_lock()) {
+    if (unlikely(mutex.try_lock())) {
         FLOGF(error, L"%s is not locked when it should be in '%s'", who, caller);
         FLOG(error, L"Break on debug_thread_error to debug.");
         debug_thread_error();
@@ -1868,7 +1837,7 @@ void redirect_tty_output() {
 
 /// Display a failed assertion message, dump a stack trace if possible, then die.
 [[noreturn]] void __fish_assert(const char *msg, const char *file, size_t line, int error) {
-    if (error) {
+    if (unlikely(error)) {
         FLOGF(error, L"%s:%zu: failed assertion: %s: errno %d (%s)", file, line, msg, error,
               std::strerror(error));
     } else {
@@ -1884,10 +1853,11 @@ bool valid_var_name_char(wchar_t chr) { return fish_iswalnum(chr) || chr == L'_'
 /// Test if the given string is a valid variable name.
 bool valid_var_name(const wcstring &str) {
     // Note do not use c_str(), we want to fail on embedded nul bytes.
-    return std::all_of(str.begin(), str.end(), valid_var_name_char);
+    return !str.empty() && std::all_of(str.begin(), str.end(), valid_var_name_char);
 }
 
 bool valid_var_name(const wchar_t *str) {
+    if (str[0] == L'\0') return false;
     for (size_t i = 0; str[i] != L'\0'; i++) {
         if (!valid_var_name_char(str[i])) return false;
     }
@@ -1898,7 +1868,9 @@ bool valid_var_name(const wchar_t *str) {
 bool valid_func_name(const wcstring &str) {
     if (str.empty()) return false;
     if (str.at(0) == L'-') return false;
+    // A function name needs to be a valid path, so no / and no NULL.
     if (str.find_first_of(L'/') != wcstring::npos) return false;
+    if (str.find_first_of(L'\0') != wcstring::npos) return false;
     return true;
 }
 
