@@ -20,82 +20,28 @@ struct callback_data_t {
     wcstring key;
 
     // The value of the variable, or none if it is erased.
-    maybe_t<wcstring> val;
+    maybe_t<env_var_t> val;
 
     /// Construct from a key and maybe a value.
-    callback_data_t(wcstring k, maybe_t<wcstring> v) : key(std::move(k)), val(std::move(v)) {}
+    callback_data_t(wcstring k, maybe_t<env_var_t> v) : key(std::move(k)), val(std::move(v)) {}
 
     /// \return whether this callback represents an erased variable.
     bool is_erase() const { return !val.has_value(); }
 };
 
-typedef std::vector<callback_data_t> callback_data_list_t;
+using callback_data_list_t = std::vector<callback_data_t>;
 
 // List of fish universal variable formats.
 // This is exposed for testing.
 enum class uvar_format_t { fish_2_x, fish_3_0, future };
 
 bool get_hostname_identifier(wcstring &result);
+
 /// Class representing universal variables.
 class env_universal_t {
-    // The table of variables. Note this is sorted; this ensures that the output file is in sorted
-    // order.
-    var_table_t vars;
-
-    // Keys that have been modified, and need to be written. A value here that is not present in
-    // vars indicates a deleted value.
-    std::unordered_set<wcstring> modified;
-
-    std::string narrow_vars_path;
-    // Path that we save to. If empty, use the default.
-    wcstring explicit_vars_path;
-
-    // A generation count which is incremented every time an exported variable is modified.
-    uint64_t export_generation{1};
-
-    // Whether it's OK to save. This may be set to false if we discover that a future version of
-    // fish wrote the uvars contents.
-    bool ok_to_save{true};
-
-    mutable std::mutex lock;
-    bool load_from_path(const std::string &path, callback_data_list_t &callbacks);
-    bool load_from_path(const wcstring &path, callback_data_list_t &callbacks);
-    void load_from_fd(int fd, callback_data_list_t &callbacks);
-
-    void set_internal(const wcstring &key, const env_var_t &var);
-    bool remove_internal(const wcstring &key);
-
-    // Functions concerned with saving.
-    bool open_and_acquire_lock(const std::string &path, autoclose_fd_t *out_fd);
-    autoclose_fd_t open_temporary_file(const wcstring &directory, wcstring *out_path);
-    bool write_to_fd(int fd, const wcstring &path);
-    bool move_new_vars_file_into_place(const wcstring &src, const wcstring &dst);
-
-    // File id from which we last read.
-    file_id_t last_read_file = kInvalidFileID;
-
-    // Given a variable table, generate callbacks representing the difference between our vars and
-    // the new vars. Also update our exports generation count as necessary.
-    void generate_callbacks_and_update_exports(const var_table_t &new_vars,
-                                               callback_data_list_t &callbacks);
-
-    // Given a variable table, copy unmodified values into self. May destructively modify
-    // vars_to_acquire.
-    void acquire_variables(var_table_t &vars_to_acquire);
-
-    static bool populate_1_variable(const wchar_t *input, env_var_t::env_var_flags_t flags,
-                                    var_table_t *vars, wcstring *storage);
-
-    static void parse_message_2x_internal(const wcstring &msg, var_table_t *vars,
-                                          wcstring *storage);
-    static void parse_message_30_internal(const wcstring &msg, var_table_t *vars,
-                                          wcstring *storage);
-    static uvar_format_t read_message_internal(int fd, var_table_t *vars);
-
-    bool save(const wcstring &directory, const wcstring &vars_path);
-
    public:
-    explicit env_universal_t(wcstring path);
+    // Construct an empty universal variables.
+    env_universal_t() = default;
 
     // Get the value of the variable with the specified name.
     maybe_t<env_var_t> get(const wcstring &name) const;
@@ -115,8 +61,14 @@ class env_universal_t {
     /// Get a view on the universal variable table.
     const var_table_t &get_table() const { return vars; }
 
-    /// Loads variables at the correct path, optionally migrating from a legacy path.
-    bool initialize(callback_data_list_t &callbacks);
+    /// Initialize this uvars for the default path, migrating legacy variables.
+    /// This should be called at most once on any given instance.
+    void initialize(callback_data_list_t &callbacks);
+
+    /// Initialize a this uvars for a given path.
+    /// This is exposed for testing only.
+    void initialize_at_path(callback_data_list_t &callbacks, wcstring path,
+                            bool migrate_legacy = false);
 
     /// Reads and writes variables at the correct path. Returns true if modified variables were
     /// written.
@@ -137,14 +89,73 @@ class env_universal_t {
     bool is_ok_to_save() const { return ok_to_save; }
 
     /// Access the export generation.
-    uint64_t get_export_generation() const;
+    uint64_t get_export_generation() const { return export_generation; }
+
+   private:
+    // Path that we save to. This is set in initialize(). If empty, initialize has not been called.
+    wcstring vars_path_;
+    std::string narrow_vars_path_;
+
+    // The table of variables.
+    var_table_t vars;
+
+    // Keys that have been modified, and need to be written. A value here that is not present in
+    // vars indicates a deleted value.
+    std::unordered_set<wcstring> modified;
+
+    // A generation count which is incremented every time an exported variable is modified.
+    uint64_t export_generation{1};
+
+    // Whether it's OK to save. This may be set to false if we discover that a future version of
+    // fish wrote the uvars contents.
+    bool ok_to_save{true};
+
+    // If true, attempt to flock the uvars file.
+    // This latches to false if the file is found to be remote, where flock may hang.
+    bool do_flock{true};
+
+    // File id from which we last read.
+    file_id_t last_read_file = kInvalidFileID;
+
+    /// \return whether we are initialized.
+    bool initialized() const { return !vars_path_.empty(); }
+
+    bool load_from_path(const wcstring &path, callback_data_list_t &callbacks);
+    bool load_from_path(const std::string &path, callback_data_list_t &callbacks);
+
+    void load_from_fd(int fd, callback_data_list_t &callbacks);
+
+    // Functions concerned with saving.
+    bool open_and_acquire_lock(const wcstring &path, autoclose_fd_t *out_fd);
+    autoclose_fd_t open_temporary_file(const wcstring &directory, wcstring *out_path);
+    bool write_to_fd(int fd, const wcstring &path);
+    bool move_new_vars_file_into_place(const wcstring &src, const wcstring &dst);
+
+    // Given a variable table, generate callbacks representing the difference between our vars and
+    // the new vars. Also update our exports generation count as necessary.
+    void generate_callbacks_and_update_exports(const var_table_t &new_vars,
+                                               callback_data_list_t &callbacks);
+
+    // Given a variable table, copy unmodified values into self.
+    void acquire_variables(var_table_t &&vars_to_acquire);
+
+    static bool populate_1_variable(const wchar_t *input, env_var_t::env_var_flags_t flags,
+                                    var_table_t *vars, wcstring *storage);
+
+    static void parse_message_2x_internal(const wcstring &msg, var_table_t *vars,
+                                          wcstring *storage);
+    static void parse_message_30_internal(const wcstring &msg, var_table_t *vars,
+                                          wcstring *storage);
+    static uvar_format_t read_message_internal(int fd, var_table_t *vars);
+
+    bool save(const wcstring &directory, const wcstring &vars_path);
 };
 
 /// The "universal notifier" is an object responsible for broadcasting and receiving universal
 /// variable change notifications. These notifications do not contain the change, but merely
 /// indicate that the uvar file has changed. It is up to the uvar subsystem to re-read the file.
 ///
-/// We support a few notificatins strategies. Not all strategies are supported on all platforms.
+/// We support a few notification strategies. Not all strategies are supported on all platforms.
 ///
 /// Notifiers may request polling, and/or provide a file descriptor to be watched for readability in
 /// select().
@@ -171,14 +182,12 @@ class universal_notifier_t {
         strategy_named_pipe,
     };
 
+    universal_notifier_t(const universal_notifier_t &) = delete;
+    universal_notifier_t &operator=(const universal_notifier_t &) = delete;
+
    protected:
     universal_notifier_t();
 
-   private:
-    // No copying.
-    universal_notifier_t &operator=(const universal_notifier_t &);
-    universal_notifier_t(const universal_notifier_t &x);
-
    public:
     static notifier_strategy_t resolve_default_strategy();
     virtual ~universal_notifier_t();
@@ -209,8 +218,4 @@ class universal_notifier_t {
 
 wcstring get_runtime_path();
 
-// Environment variable for requesting a particular universal notifier. See
-// fetch_default_strategy_from_environment for names.
-#define UNIVERSAL_NOTIFIER_ENV_NAME "fish_universal_notifier"
-
 #endif
