@@ -174,16 +174,21 @@ wcstring expand_escape_string(const wcstring &el) {
     return buff;
 }
 
+enum class parse_slice_error_t {
+    none,
+    zero_index,
+    invalid_index,
+};
+
 /// Parse an array slicing specification Returns 0 on success. If a parse error occurs, returns the
 /// index of the bad token. Note that 0 can never be a bad index because the string always starts
 /// with [.
-static size_t parse_slice(const wchar_t *in, wchar_t **end_ptr, std::vector<long> &idx,
-                          size_t array_size) {
+static size_t parse_slice(const wchar_t *const in, wchar_t **const end_ptr, std::vector<long> &idx,
+                          size_t array_size, parse_slice_error_t *const error) {
     const long size = static_cast<long>(array_size);
     size_t pos = 1;  // skip past the opening square brace
 
-    int zero_index = -1;
-    bool literal_zero_index = true;
+    *error = parse_slice_error_t::none;
 
     while (true) {
         while (iswspace(in[pos]) || (in[pos] == INTERNAL_SEPARATOR)) pos++;
@@ -192,17 +197,6 @@ static size_t parse_slice(const wchar_t *in, wchar_t **end_ptr, std::vector<long
             break;
         }
 
-        // Explicitly refuse $foo[0] as valid syntax, regardless of whether or not we're going
-        // to show an error if the index ultimately evaluates to zero. This will help newcomers
-        // to fish avoid a common off-by-one error. See #4862.
-        if (literal_zero_index) {
-            if (in[pos] == L'0') {
-                zero_index = pos;
-            } else {
-                literal_zero_index = false;
-            }
-        }
-
         const wchar_t *end;
         long tmp;
         if (idx.empty() && in[pos] == L'.' && in[pos + 1] == L'.') {
@@ -216,6 +210,13 @@ static size_t parse_slice(const wchar_t *in, wchar_t **end_ptr, std::vector<long
                 // We don't test `*end` as is typically done because we expect it to not be the null
                 // char. Ignore the case of errno==-1 because it means the end char wasn't the null
                 // char.
+                *error = parse_slice_error_t::invalid_index;
+                return pos;
+            } else if (tmp == 0) {
+                // Explicitly refuse $foo[0] as valid syntax, regardless of whether or not we're
+                // going to show an error if the index ultimately evaluates to zero. This will help
+                // newcomers to fish avoid a common off-by-one error. See #4862.
+                *error = parse_slice_error_t::zero_index;
                 return pos;
             }
         }
@@ -238,6 +239,10 @@ static size_t parse_slice(const wchar_t *in, wchar_t **end_ptr, std::vector<long
                 tmp1 = fish_wcstol(&in[pos], &end);
                 // Ignore the case of errno==-1 because it means the end char wasn't the null char.
                 if (errno > 0) {
+                    *error = parse_slice_error_t::invalid_index;
+                    return pos;
+                } else if (tmp1 == 0) {
+                    *error = parse_slice_error_t::zero_index;
                     return pos;
                 }
             }
@@ -268,14 +273,9 @@ static size_t parse_slice(const wchar_t *in, wchar_t **end_ptr, std::vector<long
             continue;
         }
 
-        literal_zero_index = literal_zero_index && tmp == 0;
         idx.push_back(i1);
     }
 
-    if (literal_zero_index && zero_index != -1) {
-        return zero_index;
-    }
-
     if (end_ptr) {
         *end_ptr = const_cast<wchar_t *>(in + pos);
     }
@@ -358,10 +358,8 @@ static expand_result_t expand_variables(wcstring instr, completion_receiver_t *o
     std::shared_ptr<history_t> history{};
     maybe_t<env_var_t> var{};
     if (var_name == L"history") {
-        // Note reader_get_history may return null, if we are running non-interactively (e.g. from
-        // web_config).
         if (is_main_thread()) {
-            history = history_t::with_name(history_session_id(env_stack_t::principal()));
+            history = history_t::with_name(history_session_id(vars));
         }
     } else if (var_name != wcstring{VARIABLE_EXPAND_EMPTY}) {
         var = vars.get(var_name);
@@ -385,14 +383,21 @@ static expand_result_t expand_variables(wcstring instr, completion_receiver_t *o
         } else if (history) {
             effective_val_count = history->size();
         }
-        size_t bad_pos =
-            parse_slice(in + slice_start, &slice_end, var_idx_list, effective_val_count);
+        parse_slice_error_t parse_error;
+        size_t bad_pos = parse_slice(in + slice_start, &slice_end, var_idx_list,
+                                     effective_val_count, &parse_error);
         if (bad_pos != 0) {
-            if (in[slice_start + bad_pos] == L'0') {
-                append_syntax_error(errors, slice_start + bad_pos,
-                                    L"array indices start at 1, not 0.");
-            } else {
-                append_syntax_error(errors, slice_start + bad_pos, L"Invalid index value");
+            switch (parse_error) {
+                case parse_slice_error_t::none:
+                    assert(false && "bad_pos != 0 but parse_slice_error_t::none!");
+                    break;
+                case parse_slice_error_t::zero_index:
+                    append_syntax_error(errors, slice_start + bad_pos,
+                                        L"array indices start at 1, not 0.");
+                    break;
+                case parse_slice_error_t::invalid_index:
+                    append_syntax_error(errors, slice_start + bad_pos, L"Invalid index value");
+                    break;
             }
             return expand_result_t::make_error(STATUS_EXPAND_ERROR);
         }
@@ -620,8 +625,9 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
     size_t paren_end = 0;
     wcstring subcmd;
 
+    bool is_quoted = false;
     switch (parse_util_locate_cmdsubst_range(input, &cursor, &subcmd, &paren_begin, &paren_end,
-                                             false)) {
+                                             false, &is_quoted)) {
         case -1: {
             append_syntax_error(errors, SOURCE_LOCATION_UNKNOWN, L"Mismatched parenthesis");
             return expand_result_t::make_error(STATUS_EXPAND_ERROR);
@@ -640,6 +646,8 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
         }
     }
 
+    bool have_dollar = paren_begin > 0 && input.at(paren_begin - 1) == L'$';
+
     wcstring_list_t sub_res;
     int subshell_status = exec_subshell_for_expand(subcmd, *ctx.parser, ctx.job_group, sub_res);
     if (subshell_status != 0) {
@@ -652,6 +660,15 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
             case STATUS_CMD_ERROR:
                 err = L"Too many active file descriptors";
                 break;
+            case STATUS_CMD_UNKNOWN:
+                err = L"Unknown command";
+                break;
+            case STATUS_ILLEGAL_CMD:
+                err = L"Commandname was invalid";
+                break;
+            case STATUS_NOT_EXECUTABLE:
+                err = L"Command not executable";
+                break;
             default:
                 err = L"Unknown error while evaluating command substitution";
                 break;
@@ -667,13 +684,21 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
         std::vector<long> slice_idx;
         const wchar_t *const slice_begin = in + tail_begin;
         wchar_t *slice_end = nullptr;
-        size_t bad_pos = parse_slice(slice_begin, &slice_end, slice_idx, sub_res.size());
+        parse_slice_error_t parse_error;
+        size_t bad_pos =
+            parse_slice(slice_begin, &slice_end, slice_idx, sub_res.size(), &parse_error);
         if (bad_pos != 0) {
-            if (slice_begin[bad_pos] == L'0') {
-                append_syntax_error(errors, slice_begin - in + bad_pos,
-                                    L"array indices start at 1, not 0.");
-            } else {
-                append_syntax_error(errors, slice_begin - in + bad_pos, L"Invalid index value");
+            switch (parse_error) {
+                case parse_slice_error_t::none:
+                    assert(false && "bad_pos != 0 but parse_slice_error_t::none!");
+                    break;
+                case parse_slice_error_t::zero_index:
+                    append_syntax_error(errors, slice_begin - in + bad_pos,
+                                        L"array indices start at 1, not 0.");
+                    break;
+                case parse_slice_error_t::invalid_index:
+                    append_syntax_error(errors, slice_begin - in + bad_pos, L"Invalid index value");
+                    break;
             }
             return expand_result_t::make_error(STATUS_EXPAND_ERROR);
         }
@@ -693,19 +718,63 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
     // Recursively call ourselves to expand any remaining command substitutions. The result of this
     // recursive call using the tail of the string is inserted into the tail_expand array list
     completion_receiver_t tail_expand_recv = out->subreceiver();
-    expand_cmdsubst(input.substr(tail_begin), ctx, &tail_expand_recv,
+    wcstring tail = input.substr(tail_begin);
+    // A command substitution inside double quotes magically closes the quoted string.
+    // Reopen the quotes just after the command substitution.
+    if (is_quoted) {
+        tail.insert(0, L"\"");
+    }
+    expand_cmdsubst(std::move(tail), ctx, &tail_expand_recv,
                     errors);  // TODO: offset error locations
     completion_list_t tail_expand = tail_expand_recv.take();
 
     // Combine the result of the current command substitution with the result of the recursive tail
     // expansion.
+
+    if (is_quoted) {
+        // Awkwardly reconstruct the command output.
+        size_t approx_size = 0;
+        for (const wcstring &sub_item : sub_res) {
+            approx_size += sub_item.size() + 1;
+        }
+        wcstring sub_res_joined;
+        sub_res_joined.reserve(approx_size);
+        for (wcstring &line : sub_res) {
+            sub_res_joined.append(escape_string_for_double_quotes(std::move(line)));
+            sub_res_joined.push_back(L'\n');
+        }
+        // Mimic POSIX shells by stripping all trailing newlines.
+        if (!sub_res_joined.empty()) {
+            size_t i;
+            for (i = sub_res_joined.size(); i > 0; i--) {
+                if (sub_res_joined[i - 1] != L'\n') break;
+            }
+            sub_res_joined.erase(i);
+        }
+        // Instead of performing cartesian product expansion, we directly insert the command
+        // substitution output into the current expansion results.
+        for (const completion_t &tail_item : tail_expand) {
+            wcstring whole_item;
+            whole_item.reserve(paren_begin + 1 + sub_res_joined.size() + 1 +
+                               tail_item.completion.size());
+            whole_item.append(input, 0, paren_begin - have_dollar);
+            whole_item.append(sub_res_joined);
+            whole_item.append(tail_item.completion.substr(const_strlen(L"\"")));
+            if (!out->add(std::move(whole_item))) {
+                return append_overflow_error(errors);
+            }
+        }
+
+        return expand_result_t::ok;
+    }
+
     for (const wcstring &sub_item : sub_res) {
         wcstring sub_item2 = escape_string(sub_item, ESCAPE_ALL);
         for (const completion_t &tail_item : tail_expand) {
             wcstring whole_item;
             whole_item.reserve(paren_begin + 1 + sub_item2.size() + 1 +
                                tail_item.completion.size());
-            whole_item.append(input, 0, paren_begin);
+            whole_item.append(input, 0, paren_begin - have_dollar);
             whole_item.push_back(INTERNAL_SEPARATOR);
             whole_item.append(sub_item2);
             whole_item.push_back(INTERNAL_SEPARATOR);
@@ -765,11 +834,8 @@ static void expand_home_directory(wcstring &input, const environment_t &vars) {
             }
         }
 
-        maybe_t<wcstring> realhome;
-        if (home) realhome = normalize_path(*home);
-
-        if (realhome) {
-            input.replace(input.begin(), input.begin() + tail_idx, *realhome);
+        if (home) {
+            input.replace(input.begin(), input.begin() + tail_idx, normalize_path(*home));
         } else {
             input[0] = L'~';
         }
@@ -933,6 +999,7 @@ expand_result_t expander_t::stage_cmdsubst(wcstring input, completion_receiver_t
     }
 }
 
+// We pass by value to match other stages. NOLINTNEXTLINE(performance-unnecessary-value-param)
 expand_result_t expander_t::stage_variables(wcstring input, completion_receiver_t *out) {
     // We accept incomplete strings here, since complete uses expand_string to expand incomplete
     // strings from the commandline.
@@ -960,9 +1027,7 @@ expand_result_t expander_t::stage_braces(wcstring input, completion_receiver_t *
 }
 
 expand_result_t expander_t::stage_home_and_self(wcstring input, completion_receiver_t *out) {
-    if (!(flags & expand_flag::skip_home_directories)) {
-        expand_home_directory(input, ctx.vars);
-    }
+    expand_home_directory(input, ctx.vars);
     expand_percent_self(input);
     if (!out->add(std::move(input))) {
         return append_overflow_error(errors);
@@ -974,7 +1039,7 @@ expand_result_t expander_t::stage_wildcards(wcstring path_to_expand, completion_
     expand_result_t result = expand_result_t::ok;
 
     remove_internal_separator(&path_to_expand, flags & expand_flag::skip_wildcards);
-    const bool has_wildcard = wildcard_has(path_to_expand, true /* internal, i.e. ANY_STRING */);
+    const bool has_wildcard = wildcard_has_internal(path_to_expand);  // e.g. ANY_STRING
     const bool for_completions = flags & expand_flag::for_completions;
     const bool skip_wildcards = flags & expand_flag::skip_wildcards;
 
@@ -1023,9 +1088,9 @@ expand_result_t expander_t::stage_wildcards(wcstring path_to_expand, completion_
                 if (auto paths_var = ctx.vars.get(for_cd ? L"CDPATH" : L"PATH")) {
                     paths = paths_var->as_list();
                 }
-                if (paths.empty()) {
-                    paths.emplace_back(for_cd ? L"." : L"");
-                }
+
+                // The current directory is always valid.
+                paths.emplace_back(for_cd ? L"." : L"");
                 for (const wcstring &next_path : paths) {
                     effective_working_dirs.push_back(
                         path_apply_working_directory(next_path, working_dir));
@@ -1132,9 +1197,7 @@ expand_result_t expander_t::expand_string(wcstring input, completion_receiver_t
 
     if (total_result == expand_result_t::ok) {
         // Hack to un-expand tildes (see #647).
-        if (!(flags & expand_flag::skip_home_directories)) {
-            unexpand_tildes(input, ctx.vars, &completions);
-        }
+        unexpand_tildes(input, ctx.vars, &completions);
         if (!out_completions->add_list(std::move(completions))) {
             total_result = append_overflow_error(errors);
         }
