@@ -50,6 +50,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 #include "history.h"
 #include "intern.h"
 #include "io.h"
+#include "parse_util.h"
 #include "parser.h"
 #include "path.h"
 #include "proc.h"
@@ -74,6 +75,8 @@ class fish_cmd_opts_t {
     std::vector<std::string> postconfig_cmds;
     /// Whether to print rusage-self stats after execution.
     bool print_rusage_self{false};
+    /// Whether no-config is set.
+    bool no_config{false};
     /// Whether no-exec is set.
     bool no_exec{false};
     /// Whether this is a login shell.
@@ -253,7 +256,26 @@ static int run_command_list(parser_t &parser, std::vector<std::string> *cmds,
                             const io_chain_t &io) {
     for (const auto &cmd : *cmds) {
         wcstring cmd_wcs = str2wcstring(cmd);
-        parser.eval(cmd_wcs, io);
+        // Parse into an ast and detect errors.
+        parse_error_list_t errors;
+        auto ast = ast::ast_t::parse(cmd_wcs, parse_flag_none, &errors);
+        bool errored = ast.errored();
+        if (!errored) {
+            errored = parse_util_detect_errors(ast, cmd_wcs, &errors);
+        }
+        if (!errored) {
+            // Construct a parsed source ref.
+            // Be careful to transfer ownership, this could be a very large string.
+            parsed_source_ref_t ps =
+                std::make_shared<parsed_source_t>(std::move(cmd_wcs), std::move(ast));
+            parser.eval(ps, io);
+        } else {
+            wcstring sb;
+            parser.get_backtrace(cmd_wcs, errors, sb);
+            std::fwprintf(stderr, L"%ls", sb.c_str());
+            // Note: We do not return here because that's how it historically worked.
+            // return 1;
+        }
     }
 
     return 0;
@@ -261,7 +283,7 @@ static int run_command_list(parser_t &parser, std::vector<std::string> *cmds,
 
 /// Parse the argument list, return the index of the first non-flag arguments.
 static int fish_parse_opt(int argc, char **argv, fish_cmd_opts_t *opts) {
-    static const char *const short_opts = "+hPilnvc:C:p:d:f:D:o:";
+    static const char *const short_opts = "+hPilNnvc:C:p:d:f:D:o:";
     static const struct option long_opts[] = {
         {"command", required_argument, nullptr, 'c'},
         {"init-command", required_argument, nullptr, 'C'},
@@ -271,6 +293,7 @@ static int fish_parse_opt(int argc, char **argv, fish_cmd_opts_t *opts) {
         {"debug-stack-frames", required_argument, nullptr, 'D'},
         {"interactive", no_argument, nullptr, 'i'},
         {"login", no_argument, nullptr, 'l'},
+        {"no-config", no_argument, nullptr, 'N'},
         {"no-execute", no_argument, nullptr, 'n'},
         {"print-rusage-self", no_argument, nullptr, 1},
         {"print-debug-categories", no_argument, nullptr, 2},
@@ -293,20 +316,10 @@ static int fish_parse_opt(int argc, char **argv, fish_cmd_opts_t *opts) {
                 break;
             }
             case 'd': {
-                char *end;
-                long tmp;
-
-                errno = 0;
-                tmp = strtol(optarg, &end, 10);
-
-                if (tmp >= 0 && tmp <= 10 && !*end && !errno) {
-                    debug_level = static_cast<int>(tmp);
-                } else {
-                    activate_flog_categories_by_pattern(str2wcstring(optarg));
-                }
+                activate_flog_categories_by_pattern(str2wcstring(optarg));
                 for (auto cat : get_flog_categories()) {
                     if (cat->enabled) {
-                        printf("Debug enabled for category: %ls\n", cat->name);
+                        std::fwprintf(stdout, L"Debug enabled for category: %ls\n", cat->name);
                     }
                 }
                 break;
@@ -331,6 +344,12 @@ static int fish_parse_opt(int argc, char **argv, fish_cmd_opts_t *opts) {
                 opts->is_login = true;
                 break;
             }
+            case 'N': {
+                opts->no_config = true;
+                // --no-config implies private mode, we won't be saving history
+                opts->enable_private_mode = true;
+                break;
+            }
             case 'n': {
                 opts->no_exec = true;
                 break;
@@ -469,7 +488,9 @@ int main(int argc, char **argv) {
     // If we're not executing, there's no need to find the config.
     if (!opts.no_exec) {
         paths = determine_config_directory_paths(argv[0]);
-        env_init(&paths);
+    }
+    if (!opts.no_exec) {
+        env_init(&paths, /* do uvars */ !opts.no_config, /* default paths */ opts.no_config);
     }
 
     // Set features early in case other initialization depends on them.
@@ -488,7 +509,7 @@ int main(int argc, char **argv) {
 
     parser_t &parser = parser_t::principal_parser();
 
-    if (!opts.no_exec) {
+    if (!opts.no_exec && !opts.no_config) {
         read_init(parser, paths);
     }
     // Re-read the terminal modes after config, it might have changed them.
@@ -498,8 +519,8 @@ int main(int argc, char **argv) {
     parser.set_last_statuses(statuses_t::just(STATUS_CMD_OK));
 
     // If we're profiling startup to a separate file, write it now.
-    if (!opts.profile_startup_output.empty()
-        && opts.profile_startup_output != opts.profile_output) {
+    if (!opts.profile_startup_output.empty() &&
+        opts.profile_startup_output != opts.profile_output) {
         parser.emit_profiling(opts.profile_startup_output.c_str());
 
         // If we are profiling both, ensure the startup data only
@@ -542,7 +563,8 @@ int main(int argc, char **argv) {
         const char *file = *(argv + (my_optind++));
         autoclose_fd_t fd(open_cloexec(file, O_RDONLY));
         if (!fd.valid()) {
-            perror(file);
+            FLOGF(error, _(L"Error reading script file '%s':"), file);
+            perror("error");
         } else {
             wcstring_list_t list;
             for (char **ptr = argv + my_optind; *ptr; ptr++) {
@@ -563,9 +585,7 @@ int main(int argc, char **argv) {
     }
 
     int exit_status = res ? STATUS_CMD_UNKNOWN : parser.get_last_status();
-
-    event_fire(parser,
-               proc_create_event(L"PROCESS_EXIT", event_type_t::exit, getpid(), exit_status));
+    event_fire(parser, event_t::process_exit(getpid(), exit_status));
 
     // Trigger any exit handlers.
     wcstring_list_t event_args = {to_string(exit_status)};
