@@ -41,6 +41,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 #include "fds.h"
 #include "fish_version.h"
 #include "flog.h"
+#include "future_feature_flags.h"
 #include "highlight.h"
 #include "operation_context.h"
 #include "output.h"
@@ -167,9 +168,10 @@ struct pretty_printer_t {
     static gap_flags_t gap_text_flags_before_node(const node_t &node) {
         gap_flags_t result = default_flags;
         switch (node.type) {
-            // Allow escaped newlines in argument and redirection lists.
+            // Allow escaped newlines before leaf nodes that can be part of a long command.
             case type_t::argument:
             case type_t::redirection:
+            case type_t::variable_assignment:
                 result |= allow_escaped_newlines;
                 break;
 
@@ -181,6 +183,23 @@ struct pretty_printer_t {
                     case parse_token_type_t::pipe:
                         result |= allow_escaped_newlines;
                         break;
+                    case parse_token_type_t::string: {
+                        // Allow escaped newlines before commands that follow a variable assignment
+                        // since both can be long (#7955).
+                        const node_t *p = node.parent;
+                        if (p->type != type_t::decorated_statement) break;
+                        p = p->parent;
+                        assert(p->type == type_t::statement);
+                        p = p->parent;
+                        if (auto job = p->try_as<job_t>()) {
+                            if (!job->variables.empty()) result |= allow_escaped_newlines;
+                        } else if (auto job_cnt = p->try_as<job_continuation_t>()) {
+                            if (!job_cnt->variables.empty()) result |= allow_escaped_newlines;
+                        } else if (auto not_stmt = p->try_as<not_statement_t>()) {
+                            if (!not_stmt->variables.empty()) result |= allow_escaped_newlines;
+                        }
+                        break;
+                    }
                     default:
                         break;
                 }
@@ -531,7 +550,28 @@ struct pretty_printer_t {
 
     template <type_t Type>
     void emit_node_text(const leaf_t<Type> &node) {
-        emit_text(node.range, gap_text_flags_before_node(node));
+        source_range_t range = node.range;
+
+        // Weird special-case: a token may end in an escaped newline. Notably, the newline is
+        // not part of the following gap text, handle indentation here (#8197).
+        bool ends_with_escaped_nl = node.range.length >= 2 &&
+                                    source.at(node.range.end() - 2) == L'\\' &&
+                                    source.at(node.range.end() - 1) == L'\n';
+        if (ends_with_escaped_nl) {
+            range = {range.start, range.length - 2};
+        }
+
+        emit_text(range, gap_text_flags_before_node(node));
+
+        if (ends_with_escaped_nl) {
+            // By convention, escaped newlines are preceded with a space.
+            output.append(L" \\\n");
+            // TODO Maybe check "allow_escaped_newlines" and use the precomputed indents.
+            // The cases where this matters are probably very rare.
+            current_indent++;
+            emit_space_or_indent();
+            current_indent--;
+        }
     }
 
     // Emit one newline.
@@ -631,6 +671,7 @@ static const char *highlight_role_to_string(highlight_role_t role) {
         TEST_ROLE(keyword)
         TEST_ROLE(statement_terminator)
         TEST_ROLE(param)
+        TEST_ROLE(option)
         TEST_ROLE(comment)
         TEST_ROLE(search_match)
         TEST_ROLE(operat)
@@ -737,6 +778,9 @@ static const wchar_t *html_class_name_for_color(highlight_spec_t spec) {
         case highlight_role_t::param: {
             return P(param);
         }
+        case highlight_role_t::option: {
+            return P(option);
+        }
         case highlight_role_t::comment: {
             return P(comment);
         }
@@ -835,6 +879,12 @@ int main(int argc, char *argv[]) {
     setlocale(LC_ALL, "");
     env_init();
 
+    if (auto features_var = env_stack_t::globals().get(L"fish_features")) {
+        for (const wcstring &s : features_var->as_list()) {
+            mutable_fish_features().set_from_string(s);
+        }
+    }
+
     // Types of output we support.
     enum {
         output_type_plain_text,
@@ -904,20 +954,10 @@ int main(int argc, char *argv[]) {
                 break;
             }
             case 'd': {
-                char *end;
-                long tmp;
-
-                errno = 0;
-                tmp = strtol(optarg, &end, 10);
-
-                if (tmp >= 0 && tmp <= 10 && !*end && !errno) {
-                    debug_level = static_cast<int>(tmp);
-                } else {
-                    activate_flog_categories_by_pattern(str2wcstring(optarg));
-                }
+                activate_flog_categories_by_pattern(str2wcstring(optarg));
                 for (auto cat : get_flog_categories()) {
                     if (cat->enabled) {
-                        printf("Debug enabled for category: %ls\n", cat->name);
+                        std::fwprintf(stdout, L"Debug enabled for category: %ls\n", cat->name);
                     }
                 }
                 break;
