@@ -135,6 +135,8 @@ static void print_ignored_signals() {
 
 static void print_stop_cont() {
     signal(SIGTSTP, [](int) {
+        // C++ compilers are awful and this is the dance we need to do to silence the "Unused
+        // result" warning. No, casting to (void) does *not* work. Please leave this.
         auto __attribute__((unused)) _ = write(STDOUT_FILENO, "SIGTSTP\n", strlen("SIGTSTP\n"));
         kill(getpid(), SIGSTOP);
     });
@@ -149,6 +151,27 @@ static void print_stop_cont() {
     }
 }
 
+static void sigkill_self() {
+    kill(getpid(), SIGKILL);
+    abort();
+}
+
+static void stdin_make_nonblocking() {
+    const int fd = STDIN_FILENO;
+    // Catch SIGCONT so pause() wakes us up.
+    signal(SIGCONT, [](int) {});
+
+    for (;;) {
+        int flags = fcntl(fd, F_GETFL, 0);
+        fprintf(stdout, "stdin was %sblocking\n", (flags & O_NONBLOCK) ? "non" : "");
+        if (fcntl(fd, F_SETFL, flags | O_NONBLOCK)) {
+            perror("fcntl");
+            exit(EXIT_FAILURE);
+        }
+        pause();
+    }
+}
+
 static void show_help();
 
 /// A thing that fish_test_helper can do.
@@ -180,6 +203,9 @@ static fth_command_t s_commands[] = {
     {"print_ignored_signals", print_ignored_signals,
      "Print to stdout the name(s) of ignored signals"},
     {"print_stop_cont", print_stop_cont, "Print when we get SIGTSTP and SIGCONT, exiting on input"},
+    {"sigkill_self", sigkill_self, "Send SIGKILL to self"},
+    {"stdin_make_nonblocking", stdin_make_nonblocking,
+     "Print if stdin is blocking and then make it nonblocking"},
     {"help", show_help, "Print list of fish_test_helper commands"},
 };
 
