@@ -37,8 +37,6 @@
 #include "wildcard.h"
 #include "wutil.h"  // IWYU pragma: keep
 
-#define CURSOR_POSITION_INVALID static_cast<size_t>(-1)
-
 static const wchar_t *get_highlight_var_name(highlight_role_t role) {
     switch (role) {
         case highlight_role_t::normal:
@@ -53,6 +51,8 @@ static const wchar_t *get_highlight_var_name(highlight_role_t role) {
             return L"fish_color_end";
         case highlight_role_t::param:
             return L"fish_color_param";
+        case highlight_role_t::option:
+            return L"fish_color_option";
         case highlight_role_t::comment:
             return L"fish_color_comment";
         case highlight_role_t::search_match:
@@ -104,59 +104,41 @@ static const wchar_t *get_highlight_var_name(highlight_role_t role) {
 static highlight_role_t get_fallback(highlight_role_t role) {
     switch (role) {
         case highlight_role_t::normal:
-            return highlight_role_t::normal;
         case highlight_role_t::error:
-            return highlight_role_t::normal;
         case highlight_role_t::command:
+        case highlight_role_t::statement_terminator:
+        case highlight_role_t::param:
+        case highlight_role_t::search_match:
+        case highlight_role_t::comment:
+        case highlight_role_t::operat:
+        case highlight_role_t::escape:
+        case highlight_role_t::quote:
+        case highlight_role_t::redirection:
+        case highlight_role_t::autosuggestion:
+        case highlight_role_t::selection:
+        case highlight_role_t::pager_progress:
+        case highlight_role_t::pager_background:
+        case highlight_role_t::pager_prefix:
+        case highlight_role_t::pager_completion:
+        case highlight_role_t::pager_description:
             return highlight_role_t::normal;
         case highlight_role_t::keyword:
             return highlight_role_t::command;
-        case highlight_role_t::statement_terminator:
-            return highlight_role_t::normal;
-        case highlight_role_t::param:
-            return highlight_role_t::normal;
-        case highlight_role_t::comment:
-            return highlight_role_t::normal;
-        case highlight_role_t::search_match:
-            return highlight_role_t::normal;
-        case highlight_role_t::operat:
-            return highlight_role_t::normal;
-        case highlight_role_t::escape:
-            return highlight_role_t::normal;
-        case highlight_role_t::quote:
-            return highlight_role_t::normal;
-        case highlight_role_t::redirection:
-            return highlight_role_t::normal;
-        case highlight_role_t::autosuggestion:
-            return highlight_role_t::normal;
-        case highlight_role_t::selection:
-            return highlight_role_t::normal;
-        case highlight_role_t::pager_progress:
-            return highlight_role_t::normal;
-        case highlight_role_t::pager_background:
-            return highlight_role_t::normal;
-        case highlight_role_t::pager_prefix:
-            return highlight_role_t::normal;
-        case highlight_role_t::pager_completion:
-            return highlight_role_t::normal;
-        case highlight_role_t::pager_description:
-            return highlight_role_t::normal;
+        case highlight_role_t::option:
+            return highlight_role_t::param;
         case highlight_role_t::pager_secondary_background:
             return highlight_role_t::pager_background;
         case highlight_role_t::pager_secondary_prefix:
+        case highlight_role_t::pager_selected_prefix:
             return highlight_role_t::pager_prefix;
         case highlight_role_t::pager_secondary_completion:
+        case highlight_role_t::pager_selected_completion:
             return highlight_role_t::pager_completion;
         case highlight_role_t::pager_secondary_description:
+        case highlight_role_t::pager_selected_description:
             return highlight_role_t::pager_description;
         case highlight_role_t::pager_selected_background:
             return highlight_role_t::search_match;
-        case highlight_role_t::pager_selected_prefix:
-            return highlight_role_t::pager_prefix;
-        case highlight_role_t::pager_selected_completion:
-            return highlight_role_t::pager_completion;
-        case highlight_role_t::pager_selected_description:
-            return highlight_role_t::pager_description;
     }
     DIE("invalid highlight role");
 }
@@ -173,7 +155,7 @@ static bool fs_is_case_insensitive(const wcstring &path, int fd,
     bool result = false;
 #ifdef _PC_CASE_SENSITIVE
     // Try the cache first.
-    case_sensitivity_cache_t::iterator cache = case_sensitivity_cache.find(path);
+    auto cache = case_sensitivity_cache.find(path);
     if (cache != case_sensitivity_cache.end()) {
         /* Use the cached value */
         result = cache->second;
@@ -322,6 +304,8 @@ static bool is_potential_cd_path(const wcstring &path, const wcstring &working_d
         auto cdpath = ctx.vars.get(L"CDPATH");
         wcstring_list_t pathsv =
             cdpath.missing_or_empty() ? wcstring_list_t{L"."} : cdpath->as_list();
+        // The current $PWD is always valid.
+        pathsv.push_back(L".");
 
         for (auto next_path : pathsv) {
             if (next_path.empty()) next_path = L".";
@@ -385,7 +369,7 @@ rgb_color_t highlight_color_resolver_t::resolve_spec(const highlight_spec_t &hig
                                                      bool is_background,
                                                      const environment_t &vars) {
     auto &cache = is_background ? bg_cache_ : fg_cache_;
-    auto p = cache.insert(std::make_pair(highlight, rgb_color_t{}));
+    auto p = cache.emplace(highlight, rgb_color_t{});
     auto iter = p.first;
     bool did_insert = p.second;
     if (did_insert) {
@@ -494,6 +478,9 @@ static size_t color_variable(const wchar_t *in, size_t in_len,
         wchar_t next = in[idx + 1];
         if (next == L'$' || valid_var_name_char(next)) {
             colors[idx] = highlight_role_t::operat;
+        } else if (next == L'(') {
+            colors[idx] = highlight_role_t::operat;
+            return idx + 1;
         } else {
             colors[idx] = highlight_role_t::error;
         }
@@ -502,26 +489,32 @@ static size_t color_variable(const wchar_t *in, size_t in_len,
     }
 
     // Handle a sequence of variable characters.
-    while (valid_var_name_char(in[idx])) {
-        colors[idx++] = highlight_role_t::operat;
+    // It may contain an escaped newline - see #8444.
+    for (;;) {
+        if (valid_var_name_char(in[idx])) {
+            colors[idx++] = highlight_role_t::operat;
+        } else if (in[idx] == L'\\' && in[idx + 1] == L'\n') {
+            colors[idx++] = highlight_role_t::operat;
+            colors[idx++] = highlight_role_t::operat;
+        } else {
+            break;
+        }
     }
 
     // Handle a slice, up to dollar_count of them. Note that we currently don't do any validation of
     // the slice's contents, e.g. $foo[blah] will not show an error even though it's invalid.
-    for (size_t slice_count = 0; slice_count < dollar_count && in[idx] == L'['; slice_count++) {
-        wchar_t *slice_begin = nullptr, *slice_end = nullptr;
-        int located = parse_util_locate_slice(in + idx, &slice_begin, &slice_end, false);
-        if (located == 1) {
-            size_t slice_begin_idx = slice_begin - in, slice_end_idx = slice_end - in;
-            assert(slice_end_idx > slice_begin_idx);
-            colors[slice_begin_idx] = highlight_role_t::operat;
-            colors[slice_end_idx] = highlight_role_t::operat;
-            idx = slice_end_idx + 1;
-        } else if (located == 0) {
+    for (size_t slice_count = 0; slice_count < dollar_count; slice_count++) {
+        long slice_len = parse_util_slice_length(in + idx);
+        if (slice_len > 0) {
+            auto slice_ulen = static_cast<size_t>(slice_len);
+            colors[idx] = highlight_role_t::operat;
+            colors[idx + slice_ulen - 1] = highlight_role_t::operat;
+            idx += slice_ulen;
+        } else if (slice_len == 0) {
             // not a slice
             break;
         } else {
-            assert(located < 0);
+            assert(slice_len < 0);
             // Syntax error. Normally the entire token is colored red for us, but inside a
             // double-quoted string that doesn't happen. As such, color the variable + the slice
             // start red. Coloring any more than that looks bad, unless we're willing to try and
@@ -538,7 +531,8 @@ static size_t color_variable(const wchar_t *in, size_t in_len,
 static void color_string_internal(const wcstring &buffstr, highlight_spec_t base_color,
                                   std::vector<highlight_spec_t>::iterator colors) {
     // Clarify what we expect.
-    assert((base_color == highlight_role_t::param || base_color == highlight_role_t::command) &&
+    assert((base_color == highlight_role_t::param || base_color == highlight_role_t::option ||
+            base_color == highlight_role_t::command) &&
            "Unexpected base color");
     const size_t buff_len = buffstr.size();
     std::fill(colors, colors + buff_len, base_color);
@@ -768,6 +762,7 @@ static void color_string_internal(const wcstring &buffstr, highlight_spec_t base
     }
 }
 
+namespace {
 /// Syntax highlighter helper.
 class highlighter_t {
     // The string we're highlighting. Note this is a reference memmber variable (to avoid copying)!
@@ -797,13 +792,11 @@ class highlighter_t {
     // Color a command.
     void color_command(const ast::string_t &node);
     // Color a node as if it were an argument.
-    void color_as_argument(const ast::node_t &node);
+    void color_as_argument(const ast::node_t &node, bool options_allowed = true);
     // Colors the source range of a node with a given color.
     void color_node(const ast::node_t &node, highlight_spec_t color);
     // Colors a range with a given color.
     void color_range(source_range_t range, highlight_spec_t color);
-    // return whether a plain statement is 'cd'.
-    bool is_cd(const ast::decorated_statement_t &stmt) const;
 
     /// \return a substring of our buffer.
     wcstring get_source(source_range_t r) const;
@@ -821,10 +814,10 @@ class highlighter_t {
     void visit(const ast::variable_assignment_t &varas);
     void visit(const ast::semi_nl_t &semi_nl);
     void visit(const ast::decorated_statement_t &stmt);
-    void visit(const ast::block_statement_t &header);
+    void visit(const ast::block_statement_t &block);
 
     // Visit an argument, perhaps knowing that our command is cd.
-    void visit(const ast::argument_t &arg, bool cmd_is_cd = false);
+    void visit(const ast::argument_t &arg, bool cmd_is_cd = false, bool options_allowed = true);
 
     // Default implementation is to just visit children.
     void visit(const ast::node_t &node) { visit_children(node); }
@@ -867,7 +860,7 @@ void highlighter_t::color_command(const ast::string_t &node) {
 }
 
 // node does not necessarily have type symbol_argument here.
-void highlighter_t::color_as_argument(const ast::node_t &node) {
+void highlighter_t::color_as_argument(const ast::node_t &node, bool options_allowed) {
     auto source_range = node.source_range();
     const wcstring arg_str = get_source(source_range);
 
@@ -876,14 +869,19 @@ void highlighter_t::color_as_argument(const ast::node_t &node) {
     const color_array_t::iterator arg_colors = color_array.begin() + arg_start;
 
     // Color this argument without concern for command substitutions.
-    color_string_internal(arg_str, highlight_role_t::param, arg_colors);
+    if (options_allowed && arg_str[0] == L'-') {
+        color_string_internal(arg_str, highlight_role_t::option, arg_colors);
+    } else {
+        color_string_internal(arg_str, highlight_role_t::param, arg_colors);
+    }
 
     // Now do command substitutions.
     size_t cmdsub_cursor = 0, cmdsub_start = 0, cmdsub_end = 0;
     wcstring cmdsub_contents;
+    bool is_quoted = false;
     while (parse_util_locate_cmdsubst_range(arg_str, &cmdsub_cursor, &cmdsub_contents,
                                             &cmdsub_start, &cmdsub_end,
-                                            true /* accept incomplete */) > 0) {
+                                            true /* accept incomplete */, &is_quoted) > 0) {
         // The cmdsub_start is the open paren. cmdsub_end is either the close paren or the end of
         // the string. cmdsub_contents extends from one past cmdsub_start to cmdsub_end.
         assert(cmdsub_end > cmdsub_start);
@@ -904,7 +902,7 @@ void highlighter_t::color_as_argument(const ast::node_t &node) {
         // Highlight it recursively.
         highlighter_t cmdsub_highlighter(cmdsub_contents, this->ctx, this->working_directory,
                                          this->io_ok);
-        const color_array_t &subcolors = cmdsub_highlighter.highlight();
+        color_array_t subcolors = cmdsub_highlighter.highlight();
 
         // Copy out the subcolors back into our array.
         assert(subcolors.size() == cmdsub_contents.size());
@@ -938,14 +936,6 @@ static bool range_is_potential_path(const wcstring &src, const source_range_t &r
     return result;
 }
 
-bool highlighter_t::is_cd(const ast::decorated_statement_t &stmt) const {
-    wcstring cmd_str;
-    if (this->io_ok && statement_get_expanded_command(this->buff, stmt, ctx, &cmd_str)) {
-        return cmd_str == L"cd";
-    }
-    return false;
-}
-
 void highlighter_t::visit(const ast::keyword_base_t &kw) {
     highlight_role_t role = highlight_role_t::normal;
     switch (kw.kw) {
@@ -1009,8 +999,8 @@ void highlighter_t::visit(const ast::semi_nl_t &semi_nl) {
     color_node(semi_nl, highlight_role_t::statement_terminator);
 }
 
-void highlighter_t::visit(const ast::argument_t &arg, bool cmd_is_cd) {
-    color_as_argument(arg);
+void highlighter_t::visit(const ast::argument_t &arg, bool cmd_is_cd, bool options_allowed) {
+    color_as_argument(arg, options_allowed);
     if (cmd_is_cd && io_ok) {
         // Mark this as an error if it's not 'help' and not a valid cd path.
         wcstring param = arg.source(this->buff);
@@ -1073,6 +1063,8 @@ void highlighter_t::visit(const ast::decorated_statement_t &stmt) {
     // Except if our command is 'cd' we have special logic for how arguments are colored.
     bool is_cd = (expanded_cmd == L"cd");
     bool is_set = (expanded_cmd == L"set");
+    // If we have seen a "--" argument, color all options from then on as normal arguments.
+    bool have_dashdash = false;
     for (const ast::argument_or_redirection_t &v : stmt.args_or_redirs) {
         if (v.is_argument()) {
             if (is_set) {
@@ -1082,7 +1074,8 @@ void highlighter_t::visit(const ast::decorated_statement_t &stmt) {
                     is_set = false;
                 }
             }
-            this->visit(v.argument(), is_cd);
+            this->visit(v.argument(), is_cd, !have_dashdash);
+            if (v.argument().source(this->buff) == L"--") have_dashdash = true;
         } else {
             this->visit(v.redirection());
         }
@@ -1240,6 +1233,57 @@ void highlighter_t::visit(const ast::redirection_t &redir) {
     }
 }
 
+highlighter_t::color_array_t highlighter_t::highlight() {
+    // If we are doing I/O, we must be in a background thread.
+    if (io_ok) {
+        ASSERT_IS_BACKGROUND_THREAD();
+    }
+
+    this->color_array.resize(this->buff.size());
+    std::fill(this->color_array.begin(), this->color_array.end(), highlight_spec_t{});
+
+    this->visit_children(*ast.top());
+    if (ctx.check_cancel()) return std::move(color_array);
+
+    // Color every comment.
+    const auto &extras = ast.extras();
+    for (const source_range_t &r : extras.comments) {
+        this->color_range(r, highlight_role_t::comment);
+    }
+
+    // Color every extra semi.
+    for (const source_range_t &r : extras.semis) {
+        this->color_range(r, highlight_role_t::statement_terminator);
+    }
+
+    // Color every error range.
+    for (const source_range_t &r : extras.errors) {
+        this->color_range(r, highlight_role_t::error);
+    }
+
+    // Underline every valid path.
+    if (io_ok) {
+        for (const ast::node_t &node : ast) {
+            const auto arg = node.try_as<ast::argument_t>();
+            if (!arg || arg->unsourced) continue;
+            if (ctx.check_cancel()) break;
+            if (range_is_potential_path(buff, arg->range, ctx, working_directory)) {
+                // Don't color highlight_role_t::error because it looks dorky. For example,
+                // trying to cd into a non-directory would show an underline and also red.
+                for (size_t i = arg->range.start, end = arg->range.start + arg->range.length;
+                     i < end; i++) {
+                    if (this->color_array.at(i).foreground != highlight_role_t::error) {
+                        this->color_array.at(i).valid_path = true;
+                    }
+                }
+            }
+        }
+    }
+
+    return std::move(color_array);
+}
+}  // namespace
+
 /// Determine if a command is valid.
 static bool command_is_valid(const wcstring &cmd, enum statement_decoration_t decoration,
                              const wcstring &working_directory, const environment_t &vars) {
@@ -1285,56 +1329,6 @@ static bool command_is_valid(const wcstring &cmd, enum statement_decoration_t de
     return is_valid;
 }
 
-highlighter_t::color_array_t highlighter_t::highlight() {
-    // If we are doing I/O, we must be in a background thread.
-    if (io_ok) {
-        ASSERT_IS_BACKGROUND_THREAD();
-    }
-
-    this->color_array.resize(this->buff.size());
-    std::fill(this->color_array.begin(), this->color_array.end(), highlight_spec_t{});
-
-    this->visit_children(*ast.top());
-    if (ctx.check_cancel()) return std::move(color_array);
-
-    // Color every comment.
-    const auto &extras = ast.extras();
-    for (const source_range_t &r : extras.comments) {
-        this->color_range(r, highlight_role_t::comment);
-    }
-
-    // Color every extra semi.
-    for (const source_range_t &r : extras.semis) {
-        this->color_range(r, highlight_role_t::statement_terminator);
-    }
-
-    // Color every error range.
-    for (const source_range_t &r : extras.errors) {
-        this->color_range(r, highlight_role_t::error);
-    }
-
-    // Underline every valid path.
-    if (io_ok) {
-        for (const ast::node_t &node : ast) {
-            const ast::argument_t *arg = node.try_as<ast::argument_t>();
-            if (!arg || arg->unsourced) continue;
-            if (ctx.check_cancel()) break;
-            if (range_is_potential_path(buff, arg->range, ctx, working_directory)) {
-                // Don't color highlight_role_t::error because it looks dorky. For example,
-                // trying to cd into a non-directory would show an underline and also red.
-                for (size_t i = arg->range.start, end = arg->range.start + arg->range.length;
-                     i < end; i++) {
-                    if (this->color_array.at(i).foreground != highlight_role_t::error) {
-                        this->color_array.at(i).valid_path = true;
-                    }
-                }
-            }
-        }
-    }
-
-    return std::move(color_array);
-}
-
 std::string colorize(const wcstring &text, const std::vector<highlight_spec_t> &colors,
                      const environment_t &vars) {
     assert(colors.size() == text.size());
