@@ -38,7 +38,7 @@ enum class separation_type_t {
 
 /// A separated_buffer_t contains a list of elements, some of which may be separated explicitly and
 /// others which must be separated further by the user (e.g. via IFS).
-class separated_buffer_t {
+class separated_buffer_t : noncopyable_t {
    public:
     struct element_t {
         std::string contents;
@@ -50,13 +50,9 @@ class separated_buffer_t {
         bool is_explicitly_separated() const { return separation == separation_type_t::explicitly; }
     };
 
-    /// separated_buffer_t may not be copied.
-    separated_buffer_t(const separated_buffer_t &) = delete;
-    void operator=(const separated_buffer_t &) = delete;
-
-    /// We may be moved.
+    /// We not be copied but may be moved.
     /// Note this leaves the moved-from value in a bogus state until clear() is called on it.
-    separated_buffer_t(separated_buffer_t &&rhs) = default;
+    separated_buffer_t(separated_buffer_t &&) = default;
     separated_buffer_t &operator=(separated_buffer_t &&) = default;
 
     /// Construct a separated_buffer_t with the given buffer limit \p limit, or 0 for no limit.
@@ -163,11 +159,7 @@ class separated_buffer_t {
 enum class io_mode_t { file, pipe, fd, close, bufferfill };
 
 /// Represents an FD redirection.
-class io_data_t {
-    // No assignment or copying allowed.
-    io_data_t(const io_data_t &rhs) = delete;
-    void operator=(const io_data_t &rhs) = delete;
-
+class io_data_t : noncopyable_t, nonmovable_t {
    protected:
     io_data_t(io_mode_t m, int fd, int source_fd) : io_mode(m), fd(fd), source_fd(source_fd) {}
 
@@ -288,7 +280,7 @@ class output_stream_t;
 /// An io_buffer_t is a buffer which can populate itself by reading from an fd.
 /// It is not an io_data_t.
 class io_buffer_t {
-public:
+   public:
     explicit io_buffer_t(size_t limit) : buffer_(limit) {}
 
     ~io_buffer_t();
@@ -358,19 +350,20 @@ class io_chain_t : public std::vector<io_data_ref_t> {
 
 /// Base class representing the output that a builtin can generate.
 /// This has various subclasses depending on the ultimate output destination.
-class output_stream_t {
+class output_stream_t : noncopyable_t, nonmovable_t {
    public:
     /// Required override point. The output stream receives a string \p s with \p amt chars.
     virtual void append(const wchar_t *s, size_t amt) = 0;
 
-    /// \return true if output was discarded. This only applies to buffered output streams.
-    virtual bool discarded() const { return false; }
-
     /// \return any internally buffered contents.
     /// This is only implemented for a string_output_stream; others flush data to their underlying
     /// receiver (fd, or separated buffer) immediately and so will return an empty string here.
     virtual const wcstring &contents() const;
 
+    /// Flush any unwritten data to the underlying device, and return an error code.
+    /// A 0 code indicates success. The base implementation returns 0.
+    virtual int flush_and_check_error();
+
     /// An optional override point. This is for explicit separation.
     virtual void append_with_separation(const wchar_t *s, size_t len, separation_type_t type);
 
@@ -388,7 +381,7 @@ class output_stream_t {
     void push_back(wchar_t c) { append(c); }
 
     // Append data from a narrow buffer, widening it.
-    void append_narrow_buffer(const separated_buffer_t &buffer);
+    void append_narrow_buffer(separated_buffer_t buffer);
 
     /// Append a format string.
     void append_format(const wchar_t *format, ...) {
@@ -400,10 +393,6 @@ class output_stream_t {
 
     void append_formatv(const wchar_t *format, va_list va) { append(vformat_string(format, va)); }
 
-    // No copying.
-    output_stream_t(const output_stream_t &s) = delete;
-    void operator=(const output_stream_t &s) = delete;
-
     output_stream_t() = default;
     virtual ~output_stream_t() = default;
 };
@@ -420,6 +409,8 @@ class fd_output_stream_t final : public output_stream_t {
     /// Construct from a file descriptor, which must be nonegative.
     explicit fd_output_stream_t(int fd) : fd_(fd) { assert(fd_ >= 0 && "Invalid fd"); }
 
+    int flush_and_check_error() override;
+
     void append(const wchar_t *s, size_t amt) override;
 
    private:
@@ -453,14 +444,14 @@ class buffered_output_stream_t final : public output_stream_t {
 
     void append(const wchar_t *s, size_t amt) override;
     void append_with_separation(const wchar_t *s, size_t len, separation_type_t type) override;
-    bool discarded() const override;
+    int flush_and_check_error() override;
 
    private:
     /// The buffer we are filling.
     std::shared_ptr<io_buffer_t> buffer_;
 };
 
-struct io_streams_t {
+struct io_streams_t : noncopyable_t {
     // Streams for out and err.
     output_stream_t &out;
     output_stream_t &err;
@@ -492,10 +483,6 @@ struct io_streams_t {
     // FIXME: this is awkwardly placed.
     std::shared_ptr<job_group_t> job_group{};
 
-    // io_streams_t cannot be copied.
-    io_streams_t(const io_streams_t &) = delete;
-    void operator=(const io_streams_t &) = delete;
-
     io_streams_t(output_stream_t &out, output_stream_t &err) : out(out), err(err) {}
 };
 
