@@ -20,7 +20,7 @@
 
 #include "ast.h"
 #include "builtin.h"
-#include "builtin_function.h"
+#include "builtins/function.h"
 #include "common.h"
 #include "complete.h"
 #include "env.h"
@@ -438,6 +438,11 @@ end_execution_reason_t parse_execution_context_t::run_for_statement(
                             FAILED_EXPANSION_VARIABLE_NAME_ERR_MSG, for_var_name.c_str());
     }
 
+    if (!valid_var_name(for_var_name)) {
+        return report_error(STATUS_INVALID_ARGS, header.var_name, BUILTIN_ERR_VARNAME, L"for",
+                            for_var_name.c_str());
+    }
+
     // Get the contents to iterate over.
     wcstring_list_t arguments;
     ast_args_list_t arg_nodes = get_argument_nodes(header.args);
@@ -447,27 +452,22 @@ end_execution_reason_t parse_execution_context_t::run_for_statement(
     }
 
     auto var = parser->vars().get(for_var_name, ENV_DEFAULT);
-    if (var && var->read_only()) {
+    if (env_var_t::flags_for(for_var_name.c_str()) & env_var_t::flag_read_only) {
         return report_error(STATUS_INVALID_ARGS, header.var_name,
-                            L"You cannot use read-only variable '%ls' in a for loop",
-                            for_var_name.c_str());
+                            L"for: %ls: cannot overwrite read-only variable", for_var_name.c_str());
     }
-    int retval;
-    if (var) {
-        retval = parser->set_var_and_fire(for_var_name, ENV_LOCAL | ENV_USER, var->as_list());
-    } else {
-        retval = parser->set_empty_var_and_fire(for_var_name, ENV_LOCAL | ENV_USER);
-    }
-    assert(retval == ENV_OK);
 
-    if (!valid_var_name(for_var_name)) {
-        return report_error(STATUS_INVALID_ARGS, header.var_name, BUILTIN_ERR_VARNAME, L"for",
-                            for_var_name.c_str());
-    }
+    auto &vars = parser->vars();
+    int retval;
+    retval = vars.set(for_var_name, ENV_LOCAL | ENV_USER, var ? var->as_list() : wcstring_list_t{});
+    assert(retval == ENV_OK);
 
     trace_if_enabled(*parser, L"for", arguments);
     block_t *fb = parser->push_block(block_t::for_block());
 
+    // We fire the same event over and over again, just construct it once.
+    event_t evt = event_t::variable_set(for_var_name);
+
     // Now drive the for loop.
     for (const wcstring &val : arguments) {
         if (auto reason = check_end_execution()) {
@@ -475,9 +475,10 @@ end_execution_reason_t parse_execution_context_t::run_for_statement(
             break;
         }
 
-        int retval = parser->set_var_and_fire(for_var_name, ENV_DEFAULT | ENV_USER, val);
+        retval = vars.set(for_var_name, ENV_DEFAULT | ENV_USER, {val});
         assert(retval == ENV_OK && "for loop variable should have been successfully set");
         (void)retval;
+        event_fire(*parser, evt);
 
         auto &ld = parser->libdata();
         ld.loop_status = loop_status_t::normals;
@@ -732,57 +733,60 @@ end_execution_reason_t parse_execution_context_t::handle_command_not_found(
     if (err_code != ENOENT) {
         return this->report_error(STATUS_NOT_EXECUTABLE, statement,
                                   _(L"The file '%ls' is not executable by this user"), cmd);
-    } else {
-        // Handle unrecognized commands with standard command not found handler that can make better
-        // error messages.
-        wcstring_list_t event_args;
-        {
-            ast_args_list_t args = get_argument_nodes(statement.args_or_redirs);
-            end_execution_reason_t arg_result =
-                this->expand_arguments_from_nodes(args, &event_args, failglob);
-
-            if (arg_result != end_execution_reason_t::ok) {
-                return arg_result;
-            }
-
-            event_args.insert(event_args.begin(), cmd_str);
-        }
-
-        wcstring buffer;
-        wcstring error;
-
-        // Redirect to stderr
-        auto io = io_chain_t{};
-        io.append_from_specs({redirection_spec_t{STDOUT_FILENO, redirection_mode_t::fd, L"2"}},
-                             L"");
-
-        if (function_exists(L"fish_command_not_found", *parser)) {
-            buffer = L"fish_command_not_found";
-            for (const wcstring &arg : event_args) {
-                buffer.push_back(L' ');
-                buffer.append(escape_string(arg, ESCAPE_ALL));
-            }
-            auto prev_statuses = parser->get_last_statuses();
-
-            event_t event(event_type_t::generic);
-            event.desc.str_param1 = L"fish_command_not_found";
-            block_t *b = parser->push_block(block_t::event_block(event));
-            parser->eval(buffer, io);
-            parser->pop_block(b);
-            parser->set_last_statuses(std::move(prev_statuses));
-        } else {
-            // If we have no handler, just print it as a normal error.
-            error = _(L"Unknown command:");
-            if (!event_args.empty()) {
-                error.push_back(L' ');
-                error.append(escape_string(event_args[0], ESCAPE_ALL));
-            }
-        }
-
-        // Here we want to report an error (so it shows a backtrace).
-        // If the handler printed text, that's already shown, so error will be empty.
-        return this->report_error(STATUS_CMD_UNKNOWN, statement, error.c_str());
     }
+
+    // Handle unrecognized commands with standard command not found handler that can make better
+    // error messages.
+    wcstring_list_t event_args;
+    {
+        ast_args_list_t args = get_argument_nodes(statement.args_or_redirs);
+        end_execution_reason_t arg_result =
+            this->expand_arguments_from_nodes(args, &event_args, failglob);
+
+        if (arg_result != end_execution_reason_t::ok) {
+            return arg_result;
+        }
+
+        event_args.insert(event_args.begin(), cmd_str);
+    }
+
+    wcstring buffer;
+    wcstring error;
+
+    // Redirect to stderr
+    auto io = io_chain_t{};
+    io.append_from_specs({redirection_spec_t{STDOUT_FILENO, redirection_mode_t::fd, L"2"}}, L"");
+
+    if (function_exists(L"fish_command_not_found", *parser)) {
+        buffer = L"fish_command_not_found";
+        for (const wcstring &arg : event_args) {
+            buffer.push_back(L' ');
+            buffer.append(escape_string(arg, ESCAPE_ALL));
+        }
+        auto prev_statuses = parser->get_last_statuses();
+
+        event_t event(event_type_t::generic);
+        event.desc.str_param1 = L"fish_command_not_found";
+        block_t *b = parser->push_block(block_t::event_block(event));
+        parser->eval(buffer, io);
+        parser->pop_block(b);
+        parser->set_last_statuses(std::move(prev_statuses));
+    } else {
+        // If we have no handler, just print it as a normal error.
+        error = _(L"Unknown command:");
+        if (!event_args.empty()) {
+            error.push_back(L' ');
+            error.append(escape_string(event_args[0], ESCAPE_ALL));
+        }
+    }
+
+    if (!cmd_str.empty() && cmd_str.at(0) == L'{') {
+        error.append(ERROR_NO_BRACE_GROUPING);
+    }
+
+    // Here we want to report an error (so it shows a backtrace).
+    // If the handler printed text, that's already shown, so error will be empty.
+    return this->report_error(STATUS_CMD_UNKNOWN, statement, error.c_str());
 }
 
 end_execution_reason_t parse_execution_context_t::expand_command(
@@ -890,8 +894,7 @@ end_execution_reason_t parse_execution_context_t::populate_plain_process(
         // from expanding the command, followed by the argument nodes themselves. E.g. if the
         // command is '$gco foo' and $gco is git checkout.
         cmd_args.push_back(cmd);
-        cmd_args.insert(cmd_args.end(), args_from_cmd_expansion.begin(),
-                        args_from_cmd_expansion.end());
+        vec_append(cmd_args, std::move(args_from_cmd_expansion));
 
         ast_args_list_t arg_nodes = get_argument_nodes(statement.args_or_redirs);
         end_execution_reason_t arg_result =
@@ -905,14 +908,11 @@ end_execution_reason_t parse_execution_context_t::populate_plain_process(
         if (reason != end_execution_reason_t::ok) {
             return reason;
         }
-
-        // Determine the process type.
-        process_type = process_type_for_command(statement, cmd);
     }
 
     // Populate the process.
     proc->type = process_type;
-    proc->set_argv(cmd_args);
+    proc->set_argv(std::move(cmd_args));
     proc->set_redirection_specs(std::move(redirections));
     proc->actual_cmd = std::move(path_to_external_command);
     return end_execution_reason_t::ok;
@@ -1339,15 +1339,17 @@ end_execution_reason_t parse_execution_context_t::run_1_job(const ast::job_t &jo
     const auto &ld = parser->libdata();
 
     auto job_control_mode = get_job_control_mode();
+    // Run all command substitutions in our pgroup.
     bool wants_job_control =
-        (job_control_mode == job_control_t::all) ||
-        ((job_control_mode == job_control_t::interactive) && parser->is_interactive()) ||
-        (ctx.job_group && ctx.job_group->wants_job_control());
+        !parser->is_command_substitution() &&
+        ((job_control_mode == job_control_t::all) ||
+         ((job_control_mode == job_control_t::interactive) && parser->is_interactive()) ||
+         (ctx.job_group && ctx.job_group->wants_job_control()));
 
     job_t::properties_t props{};
     props.initial_background = job_node.bg.has_value();
     props.skip_notification =
-        ld.is_subshell || ld.is_block || ld.is_event || !parser->is_interactive();
+        ld.is_subshell || parser->is_block() || ld.is_event || !parser->is_interactive();
     props.from_event_handler = ld.is_event;
     props.job_control = wants_job_control;
     props.wants_timing = job_node_wants_timing(job_node);
@@ -1400,7 +1402,7 @@ end_execution_reason_t parse_execution_context_t::run_1_job(const ast::job_t &jo
             remove_job(*this->parser, job.get());
         }
 
-        // Update universal variables on external conmmands.
+        // Update universal variables on external commands.
         // TODO: justify this, why not on every command?
         if (job_contained_external_command) {
             parser->vars().universal_barrier();
