@@ -26,15 +26,11 @@
 #include "parse_util.h"
 #include "proc.h"
 #include "reader.h"
-#include "sanity.h"
 #include "signal.h"
 #include "wutil.h"  // IWYU pragma: keep
 
 class io_chain_t;
 
-/// Error for evaluating in illegal scope.
-#define INVALID_SCOPE_ERR_MSG _(L"Tried to evaluate commands using invalid block type '%ls'")
-
 /// While block description.
 #define WHILE_BLOCK N_(L"'while' block")
 
@@ -97,7 +93,7 @@ parser_t::parser_t() : parser_t(env_stack_t::principal_ref()) {}
 // Out of line destructor to enable forward declaration of parse_execution_context_t
 parser_t::~parser_t() = default;
 
-std::shared_ptr<parser_t> parser_t::principal{new parser_t()};
+const std::shared_ptr<parser_t> parser_t::principal{new parser_t()};
 
 parser_t &parser_t::principal_parser() {
     ASSERT_IS_MAIN_THREAD();
@@ -105,10 +101,9 @@ parser_t &parser_t::principal_parser() {
 }
 
 int parser_t::set_var_and_fire(const wcstring &key, env_mode_flags_t mode, wcstring_list_t vals) {
-    std::vector<event_t> events;
-    int res = vars().set(key, mode, std::move(vals), &events);
-    for (const auto &evt : events) {
-        event_fire(*this, evt);
+    int res = vars().set(key, mode, std::move(vals));
+    if (res == ENV_OK) {
+        event_fire(*this, event_t::variable_set(key));
     }
     return res;
 }
@@ -123,69 +118,26 @@ int parser_t::set_empty_var_and_fire(const wcstring &key, env_mode_flags_t mode)
     return set_var_and_fire(key, mode, wcstring_list_t{});
 }
 
-// Given a new-allocated block, push it onto our block list, acquiring ownership.
 block_t *parser_t::push_block(block_t &&block) {
-    block_t new_current{block};
-    const enum block_type_t type = new_current.type();
-    new_current.src_lineno = parser_t::get_lineno();
-
-    wcstring func = new_current.function_name;
-
-    const wchar_t *filename = parser_t::current_filename();
-    if (filename != nullptr) {
-        new_current.src_filename = intern(filename);
-    }
-
-    // Types top and subst are not considered blocks for the purposes of `status is-block`.
-    if (type != block_type_t::top && type != block_type_t::subst) {
-        libdata().is_block = true;
-    }
-
-    if (type == block_type_t::breakpoint) {
-        libdata().is_breakpoint = true;
-    }
-
-    if (new_current.type() != block_type_t::top) {
-        bool shadow = (type == block_type_t::function_call);
-        vars().push(shadow);
-        new_current.wants_pop_env = true;
+    block.src_lineno = parser_t::get_lineno();
+    block.src_filename = parser_t::current_filename();
+    if (block.type() != block_type_t::top) {
+        bool new_scope = (block.type() == block_type_t::function_call);
+        vars().push(new_scope);
+        block.wants_pop_env = true;
     }
 
     // Push it onto our list and return a pointer to it.
     // Note that deques do not move their contents so this is safe.
-    this->block_list.push_front(std::move(new_current));
+    this->block_list.push_front(std::move(block));
     return &this->block_list.front();
 }
 
 void parser_t::pop_block(const block_t *expected) {
-    assert(expected == this->current_block());
-    assert(!block_list.empty() && "empty block list");
-
-    // Acquire ownership out of the block list.
-    block_t old = block_list.front();
-    block_list.pop_front();
-
-    if (old.wants_pop_env) vars().pop();
-
-    // Figure out if `status is-block` should consider us to be in a block now.
-    bool new_is_block = false;
-    for (const auto &b : block_list) {
-        if (b.type() != block_type_t::top && b.type() != block_type_t::subst) {
-            new_is_block = true;
-            break;
-        }
-    }
-    libdata().is_block = new_is_block;
-
-    // Are we still in a breakpoint?
-    bool new_is_breakpoint = false;
-    for (const auto &b : block_list) {
-        if (b.type() == block_type_t::breakpoint) {
-            new_is_breakpoint = true;
-            break;
-        }
-    }
-    libdata().is_breakpoint = new_is_breakpoint;
+    assert(expected && expected == &this->block_list.at(0) && "Unexpected block");
+    bool pop_env = expected->wants_pop_env;
+    block_list.pop_front();  // beware, this deallocates 'expected'.
+    if (pop_env) vars().pop();
 }
 
 const wchar_t *parser_t::get_block_desc(block_type_t block) {
@@ -229,8 +181,6 @@ block_t *parser_t::block_at_index(size_t idx) {
     return idx < block_list.size() ? &block_list[idx] : nullptr;
 }
 
-block_t *parser_t::current_block() { return block_at_index(0); }
-
 /// Print profiling information to the specified stream.
 static void print_profile(const std::deque<profile_item_t> &items, FILE *out) {
     for (size_t idx = 0; idx < items.size(); idx++) {
@@ -272,9 +222,7 @@ static void print_profile(const std::deque<profile_item_t> &items, FILE *out) {
     }
 }
 
-void parser_t::clear_profiling() {
-    profile_items.clear();
-}
+void parser_t::clear_profiling() { profile_items.clear(); }
 
 void parser_t::emit_profiling(const char *path) const {
     // Save profiling information. OK to not use CLO_EXEC here because this is called while fish is
@@ -414,30 +362,50 @@ wcstring parser_t::stack_trace() const {
     return trace;
 }
 
-/// Returns the name of the currently evaluated function if we are currently evaluating a function,
-/// NULL otherwise. This is tested by moving down the block-scope-stack, checking every block if it
-/// is of type FUNCTION_CALL. If the caller doesn't specify a starting position in the stack we
-/// begin with the current block.
-const wchar_t *parser_t::is_function(size_t idx) const {
-    // PCA: Have to make this a string somehow.
-    ASSERT_IS_MAIN_THREAD();
-
-    for (size_t block_idx = idx; block_idx < block_list.size(); block_idx++) {
-        const block_t &b = block_list[block_idx];
+bool parser_t::is_function() const {
+    for (const auto &b : block_list) {
         if (b.is_function_call()) {
-            return b.function_name.c_str();
+            return true;
         } else if (b.type() == block_type_t::source) {
-            // If a function sources a file, obviously that function's offset doesn't
-            // contribute.
+            // If a function sources a file, don't descend further.
             break;
         }
     }
-    return nullptr;
+    return false;
 }
 
-/// Return the function name for the specified stack frame. Default is zero (current frame).
-/// The special value zero means the function frame immediately above the closest breakpoint frame.
-const wchar_t *parser_t::get_function_name(int level) {
+bool parser_t::is_block() const {
+    // Note historically this has descended into 'source', unlike 'is_function'.
+    for (const auto &b : block_list) {
+        if (b.type() != block_type_t::top && b.type() != block_type_t::subst) {
+            return true;
+        }
+    }
+    return false;
+}
+
+bool parser_t::is_breakpoint() const {
+    for (const auto &b : block_list) {
+        if (b.type() == block_type_t::breakpoint) {
+            return true;
+        }
+    }
+    return false;
+}
+
+bool parser_t::is_command_substitution() const {
+    for (const auto &b : block_list) {
+        if (b.type() == block_type_t::subst) {
+            return true;
+        } else if (b.type() == block_type_t::source) {
+            // If a function sources a file, don't descend further.
+            break;
+        }
+    }
+    return false;
+}
+
+maybe_t<wcstring> parser_t::get_function_name(int level) {
     if (level == 0) {
         // Return the function name for the level preceding the most recent breakpoint. If there
         // isn't one return the function name for the current level.
@@ -447,13 +415,10 @@ const wchar_t *parser_t::get_function_name(int level) {
             if (b.type() == block_type_t::breakpoint) {
                 found_breakpoint = true;
             } else if (found_breakpoint && b.is_function_call()) {
-                return b.function_name.c_str();
+                return b.function_name;
             }
         }
-        return nullptr;  // couldn't find a breakpoint frame
-    } else if (level == 1) {
-        // Return the function name for the current level.
-        return this->is_function();
+        return none();  // couldn't find a breakpoint frame
     }
 
     // Level 1 is the topmost function call. Level 2 is its caller. Etc.
@@ -462,11 +427,15 @@ const wchar_t *parser_t::get_function_name(int level) {
         if (b.is_function_call()) {
             funcs_seen++;
             if (funcs_seen == level) {
-                return b.function_name.c_str();
+                return b.function_name;
             }
+        } else if (b.type() == block_type_t::source && level == 1) {
+            // Historical: If we want the topmost function, but we are really in a file sourced by a
+            // function, don't consider ourselves to be in a function.
+            break;
         }
     }
-    return nullptr;  // couldn't find that function level
+    return none();
 }
 
 int parser_t::get_lineno() const {
@@ -478,11 +447,10 @@ int parser_t::get_lineno() const {
 }
 
 const wchar_t *parser_t::current_filename() const {
-    ASSERT_IS_MAIN_THREAD();
-
     for (const auto &b : block_list) {
         if (b.is_function_call()) {
-            return function_get_definition_file(b.function_name);
+            auto props = function_get_props(b.function_name);
+            return props ? props->definition_file : nullptr;
         } else if (b.type() == block_type_t::source) {
             return b.sourced_file;
         }
@@ -568,16 +536,16 @@ void parser_t::job_promote(job_t *job) {
     std::rotate(job_list.begin(), loc, std::next(loc));
 }
 
-job_t *parser_t::job_get(job_id_t id) {
+const job_t *parser_t::job_with_id(job_id_t id) const {
     for (const auto &job : job_list) {
         if (id <= 0 || job->job_id() == id) return job.get();
     }
     return nullptr;
 }
 
-const job_t *parser_t::job_get(job_id_t id) const {
+const job_t *parser_t::job_with_internal_id(internal_job_id_t id) const {
     for (const auto &job : job_list) {
-        if (id <= 0 || job->job_id() == id) return job.get();
+        if (job->internal_job_id == id) return job.get();
     }
     return nullptr;
 }
