@@ -31,6 +31,7 @@
 #endif
 #include <sys/time.h>  // IWYU pragma: keep
 #include <sys/types.h>
+#include <sys/wait.h>
 
 #include <algorithm>  // IWYU pragma: keep
 #include <memory>
@@ -49,7 +50,6 @@
 #include "parser.h"
 #include "proc.h"
 #include "reader.h"
-#include "sanity.h"
 #include "signal.h"
 #include "wcstringutil.h"
 #include "wutil.h"  // IWYU pragma: keep
@@ -95,17 +95,20 @@ void set_job_control_mode(job_control_t mode) {
 
 void proc_init() { signal_set_handlers_once(false); }
 
-/// Return true if all processes in the job have stopped or completed.
+/// Return true if all processes in the job are stopped or completed, and there is at least one
+/// stopped process.
 bool job_t::is_stopped() const {
+    bool has_stopped = false;
     for (const process_ptr_t &p : processes) {
         if (!p->completed && !p->stopped) {
             return false;
         }
+        has_stopped |= p->stopped;
     }
-    return true;
+    return has_stopped;
 }
 
-/// Return true if the last processes in the job has completed.
+/// Return true if all processes in the job have completed.
 bool job_t::is_completed() const {
     assert(!processes.empty());
     for (const process_ptr_t &p : processes) {
@@ -116,29 +119,16 @@ bool job_t::is_completed() const {
     return true;
 }
 
-bool job_t::should_report_process_exits() const {
-    // This implements the behavior of process exit events only being sent for jobs containing an
-    // external process. Bizarrely the process exit event is for the pgroup leader which may be fish
-    // itself.
-    // TODO: rationalize this.
+bool job_t::posts_job_exit_events() const {
     // If we never got a pgid then we never launched the external process, so don't report it.
-    if (!this->get_pgid()) {
-        return false;
-    }
+    if (!this->get_pgid()) return false;
 
     // Only report root job exits.
     // For example in `ls | begin ; cat ; end` we don't need to report the cat sub-job.
-    if (!flags().is_group_root) {
-        return false;
-    }
+    if (!flags().is_group_root) return false;
 
-    // Return whether we have an external process.
-    for (const auto &p : this->processes) {
-        if (p->type == process_type_t::external) {
-            return true;
-        }
-    }
-    return false;
+    // Only jobs with external processes post job_exit events.
+    return this->has_external_proc();
 }
 
 bool job_t::job_chain_is_fully_constructed() const { return group->is_root_constructed(); }
@@ -271,8 +261,12 @@ void process_t::check_generations_before_launch() {
 }
 
 void process_t::mark_aborted_before_launch() {
-    completed = true;
-    status = proc_status_t::from_exit_code(EXIT_FAILURE);
+    this->completed = true;
+    // The status may have already been set to e.g. STATUS_NOT_EXECUTABLE.
+    // Only stomp a successful status.
+    if (this->status.is_success()) {
+        this->status = proc_status_t::from_exit_code(EXIT_FAILURE);
+    }
 }
 
 bool process_t::is_internal() const {
@@ -295,6 +289,17 @@ bool process_t::is_internal() const {
     return true;
 }
 
+wait_handle_ref_t process_t::make_wait_handle(internal_job_id_t jid) {
+    if (type != process_type_t::external || pid <= 0) {
+        // Not waitable.
+        return nullptr;
+    }
+    if (!wait_handle_) {
+        wait_handle_ = std::make_shared<wait_handle_t>(this->pid, jid, wbasename(this->actual_cmd));
+    }
+    return wait_handle_;
+}
+
 static uint64_t next_internal_job_id() {
     static std::atomic<uint64_t> s_next{};
     return ++s_next;
@@ -333,8 +338,8 @@ bool job_t::has_external_proc() const {
 /// we exit. Poll these from time-to-time to prevent zombie processes from happening (#5342).
 static owning_lock<std::vector<pid_t>> s_disowned_pids;
 
-void add_disowned_job(job_t *j) {
-    if (j == nullptr) return;
+void add_disowned_job(const job_t *j) {
+    assert(j && "Null job");
 
     // Never add our own (or an invalid) pgid as it is not unique to only
     // one job, and may result in a deadlock if we attempt the wait.
@@ -434,7 +439,7 @@ static void process_mark_finished_children(parser_t &parser, bool block_ok) {
                 j->group->set_is_foreground(false);
             }
             if (status.continued()) {
-                j->mut_flags().notified = false;
+                j->mut_flags().notified_of_stop = false;
             }
             if (status.normal_exited() || status.signal_exited()) {
                 FLOGF(proc_reap_external, "Reaped external process '%ls' (pid %d, status %d)",
@@ -475,46 +480,148 @@ static void process_mark_finished_children(parser_t &parser, bool block_ok) {
     reap_disowned_pids();
 }
 
-/// Call the fish_job_summary function with the given args.
-static void print_job_summary(parser_t &parser, const wcstring_list_t &args) {
-    wcstring buffer = wcstring(L"fish_job_summary");
-    for (const wcstring &arg : args) {
-        buffer.push_back(L' ');
-        buffer.append(escape_string(arg, ESCAPE_ALL));
+/// Given a job that has completed, generate job_exit, process_exit, and caller_exit events.
+static void generate_exit_events(const job_ref_t &j, const enum_set_t<event_type_t> handled,
+                                 std::vector<event_t> *out_evts) {
+    // Generate proc and job exit events, except for jobs originating in event handlers.
+    if (handled.get(event_type_t::process_exit) && !j->from_event_handler()) {
+        for (const auto &proc : j->processes) {
+            if (proc->pid > 0) {
+                out_evts->push_back(event_t::process_exit(proc->pid, proc->status.status_value()));
+            }
+        }
     }
+
+    if (handled.get(event_type_t::job_exit) && !j->from_event_handler() &&
+        j->posts_job_exit_events()) {
+        if (auto last_pid = j->get_last_pid()) {
+            out_evts->push_back(event_t::job_exit(*last_pid, j->internal_job_id));
+        }
+    }
+
+    // Generate caller_exit events.
+    if (handled.get(event_type_t::caller_exit)) {
+        out_evts->push_back(event_t::caller_exit(j->internal_job_id, j->job_id()));
+    }
+}
+
+/// \return whether to emit a fish_job_summary call for a process.
+static bool proc_wants_summary(const shared_ptr<job_t> &j, const process_ptr_t &p) {
+    // Are we completed with a pid?
+    if (!p->completed || !p->pid) return false;
+
+    // Did we die due to a signal other than SIGPIPE?
+    auto s = p->status;
+    if (!s.signal_exited() || s.signal_code() == SIGPIPE) return false;
+
+    // Does the job want to suppress notifications?
+    // Note we always report crashes.
+    if (j->skip_notification() && !contains(crashsignals, s.signal_code())) return false;
+
+    return true;
+}
+
+/// \return whether to emit a fish_job_summary call for a job as a whole. We may also emit this for
+/// its individual processes.
+static bool job_wants_summary(const shared_ptr<job_t> &j) {
+    // Do we just skip notifications?
+    if (j->skip_notification()) return false;
+
+    // Do we have a single process which will also report? If so then that suffices for us.
+    if (j->processes.size() == 1 && proc_wants_summary(j, j->processes.front())) return false;
+
+    // Are we foreground?
+    // The idea here is to not print status messages for jobs that execute in the foreground (i.e.
+    // without & and without being `bg`).
+    if (j->is_foreground()) return false;
+
+    return true;
+}
+
+/// \return whether we want to emit a fish_job_summary call for a job or any of its processes.
+bool job_or_proc_wants_summary(const shared_ptr<job_t> &j) {
+    if (job_wants_summary(j)) return true;
+    for (const auto &p : j->processes) {
+        if (proc_wants_summary(j, p)) return true;
+    }
+    return false;
+}
+
+/// Invoke the fish_job_summary function by executing the given command.
+static void call_job_summary(parser_t &parser, const wcstring &cmd) {
     event_t event(event_type_t::generic);
     event.desc.str_param1 = L"fish_job_summary";
-
-    auto prev_statuses = parser.get_last_statuses();
-
     block_t *b = parser.push_block(block_t::event_block(event));
-    parser.eval(buffer, io_chain_t());
+    auto saved_status = parser.get_last_statuses();
+    parser.eval(cmd, io_chain_t());
+    parser.set_last_statuses(saved_status);
     parser.pop_block(b);
-
-    parser.set_last_statuses(std::move(prev_statuses));
 }
 
-/// Format information about job status for the user to look at.
-using job_status_t = enum { JOB_STOPPED, JOB_ENDED };
-static void print_job_status(parser_t &parser, const job_t *j, job_status_t status) {
-    wcstring_list_t args = {
-        format_string(L"%d", j->job_id()),
-        format_string(L"%d", j->is_foreground()),
-        j->command(),
-        status == JOB_STOPPED ? L"STOPPED" : L"ENDED",
-    };
-    print_job_summary(parser, args);
+// \return a command which invokes fish_job_summary.
+// The process pointer may be null, in which case it represents the entire job.
+// Note this implements the arguments which fish_job_summary expects.
+wcstring summary_command(const job_ref_t &j, const process_ptr_t &p = nullptr) {
+    wcstring buffer = L"fish_job_summary";
+
+    // Job id.
+    append_format(buffer, L" %d", j->job_id());
+
+    // 1 if foreground, 0 if background.
+    append_format(buffer, L" %d", static_cast<int>(j->is_foreground()));
+
+    // Command.
+    buffer.push_back(L' ');
+    buffer.append(escape_string(j->command(), ESCAPE_ALL));
+
+    if (!p) {
+        // No process, we are summarizing the whole job.
+        buffer.append(j->is_stopped() ? L" STOPPED" : L" ENDED");
+    } else {
+        // We are summarizing a process which exited with a signal.
+        // Arguments are the signal name and description.
+        int sig = p->status.signal_code();
+        buffer.push_back(L' ');
+        buffer.append(escape_string(sig2wcs(sig), ESCAPE_ALL));
+
+        buffer.push_back(L' ');
+        buffer.append(escape_string(signal_get_desc(sig), ESCAPE_ALL));
+
+        // If we have multiple processes, we also append the pid and argv.
+        if (j->processes.size() > 1) {
+            append_format(buffer, L" %d", p->pid);
+
+            buffer.push_back(L' ');
+            buffer.append(escape_string(p->argv0(), ESCAPE_ALL));
+        }
+    }
+    return buffer;
 }
 
-event_t proc_create_event(const wchar_t *msg, event_type_t type, pid_t pid, int status) {
-    event_t event{type};
-    event.desc.param1.pid = pid;
+// Summarize a list of jobs, by emitting calls to fish_job_summary.
+// Note the given list must NOT be the parser's own job list, since the call to fish_job_summary
+// could modify it.
+static bool summarize_jobs(parser_t &parser, const std::vector<job_ref_t> &jobs) {
+    if (jobs.empty()) return false;
 
-    event.arguments.reserve(3);
-    event.arguments.push_back(msg);
-    event.arguments.push_back(to_string(pid));
-    event.arguments.push_back(to_string(status));
-    return event;
+    for (const auto &j : jobs) {
+        if (j->is_stopped()) {
+            call_job_summary(parser, summary_command(j));
+        } else {
+            // Completed job.
+            for (const auto &p : j->processes) {
+                if (proc_wants_summary(j, p)) {
+                    call_job_summary(parser, summary_command(j, p));
+                }
+            }
+
+            // Overall status for the job.
+            if (job_wants_summary(j)) {
+                call_job_summary(parser, summary_command(j));
+            }
+        }
+    }
+    return true;
 }
 
 /// Remove all disowned jobs whose job chain is fully constructed (that is, do not erase disowned
@@ -531,84 +638,38 @@ static void remove_disowned_jobs(job_list_t &jobs) {
     }
 }
 
-/// Given a a process in a job, print the status message for the process as appropriate, and then
-/// mark the status code so we don't print again. Populate any events into \p exit_events.
-/// \return true if we printed a status message, false if not.
-static bool try_clean_process_in_job(parser_t &parser, process_t *p, job_t *j,
-                                     std::vector<event_t> *exit_events) {
-    if (!p->completed || !p->pid) {
-        return false;
+/// Given that a job has completed, check if it may be wait'ed on; if so add it to the wait handle
+/// store. Then mark all wait handles as complete.
+static void save_wait_handle_for_completed_job(const shared_ptr<job_t> &job,
+                                               wait_handle_store_t &store) {
+    assert(job && job->is_completed() && "Job null or not completed");
+    // Are we a background job?
+    if (!job->is_foreground()) {
+        for (auto &proc : job->processes) {
+            store.add(proc->make_wait_handle(job->internal_job_id));
+        }
     }
 
-    auto s = p->status;
-
-    // Add an exit event if the process did not come from a job handler.
-    if (!j->from_event_handler()) {
-        exit_events->push_back(proc_create_event(L"PROCESS_EXIT", event_type_t::exit, p->pid,
-                                                 s.normal_exited() ? s.exit_code() : -1));
+    // Mark all wait handles as complete (but don't create just for this).
+    for (auto &proc : job->processes) {
+        if (wait_handle_ref_t wh = proc->get_wait_handle()) {
+            wh->status = proc->status.status_value();
+            wh->completed = true;
+        }
     }
-
-    // Ignore SIGPIPE. We issue it ourselves to the pipe writer when the pipe reader dies.
-    if (!s.signal_exited() || s.signal_code() == SIGPIPE) {
-        return false;
-    }
-
-    int proc_is_job = (p->is_first_in_job && p->is_last_in_job);
-    if (proc_is_job) j->mut_flags().notified = true;
-
-    // Handle signals other than SIGPIPE.
-    // Always report crashes.
-    if (j->skip_notification() && !contains(crashsignals, s.signal_code())) {
-        return false;
-    }
-
-    wcstring_list_t args;
-    args.reserve(proc_is_job ? 5 : 7);
-    args.push_back(format_string(L"%d", j->job_id()));
-    args.push_back(format_string(L"%d", j->is_foreground()));
-    args.push_back(j->command());
-    args.push_back(sig2wcs(s.signal_code()));
-    args.push_back(signal_get_desc(s.signal_code()));
-    if (!proc_is_job) {
-        args.push_back(format_string(L"%d", p->pid));
-        args.push_back(p->argv0());
-    }
-    print_job_summary(parser, args);
-    // Clear status so it is not reported more than once.
-    // TODO: this seems like a clumsy way to ensure that.
-    p->status = proc_status_t::from_exit_code(0);
-    return true;
-}
-
-/// \return whether this job wants a status message printed when it stops or completes.
-static bool job_wants_message(const shared_ptr<job_t> &j) {
-    // Did we already print a status message?
-    if (j->flags().notified) return false;
-
-    // Do we just skip notifications?
-    if (j->skip_notification()) return false;
-
-    // Are we foreground?
-    // The idea here is to not print status messages for jobs that execute in the foreground (i.e.
-    // without & and without being `bg`).
-    if (j->is_foreground()) return false;
-
-    return true;
 }
 
 /// Remove completed jobs from the job list, printing status messages as appropriate.
 /// \return whether something was printed.
 static bool process_clean_after_marking(parser_t &parser, bool allow_interactive) {
     ASSERT_IS_MAIN_THREAD();
-    bool printed = false;
 
     // This function may fire an event handler, we do not want to call ourselves recursively (to
     // avoid infinite recursion).
     if (parser.libdata().is_cleaning_procs) {
         return false;
     }
-    parser.libdata().is_cleaning_procs = true;
-    const cleanup_t cleanup([&] { parser.libdata().is_cleaning_procs = false; });
+    const scoped_push<bool> cleaning(&parser.libdata().is_cleaning_procs, true);
 
     // This may be invoked in an exit handler, after the TERM has been torn down
     // Don't try to print in that case (#3222)
@@ -621,63 +682,62 @@ static bool process_clean_after_marking(parser_t &parser, bool allow_interactive
     // complete.
     std::vector<event_t> exit_events;
 
-    // A helper to indicate if we should process a job.
+    // Figure out which event types have handlers, so we can avoid the cost of creating events that
+    // are unhandled. In principle, an event handler could add a handler for another event type,
+    // which won't receive it. We ignore that possibility.
+    const enum_set_t<event_type_t> handled_event_types = event_get_handled_types();
+
+    // Defer processing under-construction jobs or jobs that want a message when we are not
+    // interactive.
     auto should_process_job = [=](const shared_ptr<job_t> &j) {
         // Do not attempt to process jobs which are not yet constructed.
         // Do not attempt to process jobs that need to print a status message,
         // unless we are interactive, in which case printing is OK.
-        return j->is_constructed() && (interactive || !job_wants_message(j));
+        return j->is_constructed() && (interactive || !job_or_proc_wants_summary(j));
     };
 
-    // Print status messages for completed or stopped jobs.
+    // The list of jobs to summarize. Some of these jobs are completed and are removed from the
+    // parser's job list, others are stopped and remain in the list.
+    std::vector<job_ref_t> jobs_to_summarize;
+
+    // Handle stopped jobs. These stay in our list.
     for (const auto &j : parser.jobs()) {
-        if (!should_process_job(j)) continue;
-
-        // Clean processes within the job.
-        // Note this may print the message on behalf of the job, affecting the result of
-        // job_wants_message().
-        for (process_ptr_t &p : j->processes) {
-            if (try_clean_process_in_job(parser, p.get(), j.get(), &exit_events)) {
-                printed = true;
-            }
-        }
-
-        // Print the message if we need to.
-        if (job_wants_message(j) && (j->is_completed() || j->is_stopped())) {
-            print_job_status(parser, j.get(), j->is_completed() ? JOB_ENDED : JOB_STOPPED);
-            j->mut_flags().notified = true;
-            printed = true;
-        }
-
-        // Prepare events for completed jobs
-        if (j->is_completed()) {
-            // If this job already came from an event handler,
-            // don't create an event or it's easy to get an infinite loop.
-            if (!j->from_event_handler() && j->should_report_process_exits()) {
-                pid_t pgid = *j->get_pgid();
-                exit_events.push_back(proc_create_event(L"JOB_EXIT", event_type_t::exit, -pgid, 0));
-            }
-            // Caller exit events we still create, which anecdotally fixes `source (thing | psub)` inside event handlers.
-            // This seems benign since this event is barely used (basically only psub), and it seems hard
-            // to construct an infinite loop with it.
-            exit_events.push_back(
-                proc_create_event(L"JOB_EXIT", event_type_t::caller_exit, j->job_id(), 0));
-            exit_events.back().desc.param1.caller_id = j->internal_job_id;
+        if (j->is_stopped() && !j->flags().notified_of_stop && should_process_job(j) &&
+            job_wants_summary(j)) {
+            j->mut_flags().notified_of_stop = true;
+            jobs_to_summarize.push_back(j);
         }
     }
 
-    // Remove completed jobs.
-    // Do this before calling out to user code in the event handler below, to ensure an event
-    // handler doesn't remove jobs on our behalf.
-    auto should_remove = [&](const shared_ptr<job_t> &j) {
-        return should_process_job(j) && j->is_completed();
-    };
-    auto &jobs = parser.jobs();
-    jobs.erase(std::remove_if(jobs.begin(), jobs.end(), should_remove), jobs.end());
+    // Remove completed, processable jobs from our job list.
+    for (auto iter = parser.jobs().begin(); iter != parser.jobs().end();) {
+        const job_ref_t &j = *iter;
+        if (!should_process_job(j) || !j->is_completed()) {
+            ++iter;
+            continue;
+        }
+        // We are committed to removing this job.
+        // Remember it for summary later, generate exit events, maybe save its wait handle if it
+        // finished in the background.
+        if (job_or_proc_wants_summary(j)) jobs_to_summarize.push_back(j);
+        generate_exit_events(j, handled_event_types, &exit_events);
+        save_wait_handle_for_completed_job(j, parser.get_wait_handles());
+
+        // Remove it.
+        iter = parser.jobs().erase(iter);
+    }
+
+    // Emit calls to fish_job_summary.
+    bool printed = summarize_jobs(parser, jobs_to_summarize);
 
     // Post pending exit events.
-    for (const auto &evt : exit_events) {
-        event_fire(parser, evt);
+    // If we have none, still give any other pending events (e.g. from signals) a chance to run.
+    if (exit_events.empty()) {
+        event_fire_delayed(parser);
+    } else {
+        for (const auto &evt : exit_events) {
+            event_fire(parser, evt);
+        }
     }
 
     if (printed) {
@@ -689,23 +749,18 @@ static bool process_clean_after_marking(parser_t &parser, bool allow_interactive
 
 bool job_reap(parser_t &parser, bool allow_interactive) {
     ASSERT_IS_MAIN_THREAD();
-    process_mark_finished_children(parser, false);
+    // Early out for the common case that there are no jobs.
+    if (parser.jobs().empty()) {
+        return false;
+    }
 
-    // Preserve the exit status.
-    auto saved_statuses = parser.get_last_statuses();
-
-    bool printed = process_clean_after_marking(parser, allow_interactive);
-
-    // Restore the exit status.
-    parser.set_last_statuses(std::move(saved_statuses));
-
-    return printed;
+    process_mark_finished_children(parser, false /* not block_ok */);
+    return process_clean_after_marking(parser, allow_interactive);
 }
 
 /// Get the CPU time for the specified process.
-unsigned long proc_get_jiffies(process_t *p) {
-    if (!have_proc_stat()) return 0;
-    if (p->pid <= 0) return 0;
+unsigned long proc_get_jiffies(pid_t inpid) {
+    if (inpid <= 0 || !have_proc_stat()) return 0;
 
     char state;
     int pid, ppid, pgrp, session, tty_nr, tpgid, exit_signal, processor;
@@ -718,7 +773,7 @@ unsigned long proc_get_jiffies(process_t *p) {
     /// Maximum length of a /proc/[PID]/stat filename.
     constexpr size_t FN_SIZE = 256;
     char fn[FN_SIZE];
-    std::snprintf(fn, FN_SIZE, "/proc/%d/stat", p->pid);
+    std::snprintf(fn, FN_SIZE, "/proc/%d/stat", inpid);
     // Don't use autoclose_fd here, we will fdopen() and then fclose() instead.
     int fd = open_cloexec(fn, O_RDONLY);
     if (fd < 0) return 0;
@@ -745,8 +800,8 @@ unsigned long proc_get_jiffies(process_t *p) {
 void proc_update_jiffies(parser_t &parser) {
     for (const auto &job : parser.jobs()) {
         for (process_ptr_t &p : job->processes) {
-            gettimeofday(&p->last_time, nullptr);
-            p->last_jiffies = proc_get_jiffies(p.get());
+            p->last_time = timef();
+            p->last_jiffies = proc_get_jiffies(p->pid);
         }
     }
 }
@@ -869,7 +924,7 @@ int terminal_maybe_give_to_job_group(const job_group_t *jg, bool continuing_from
             } else {
                 // Debug the original tcsetpgrp error (not the waitpid errno) to the log, and
                 // then retry until not EPERM or the process group has exited.
-                FLOGF(proc_termowner, L"terminal_give_to_job(): EPERM.\n", pgid);
+                FLOGF(proc_termowner, L"terminal_give_to_job(): EPERM with pgid %d.", pgid);
                 continue;
             }
         } else if (errno == ENOTTY) {
@@ -934,12 +989,20 @@ bool job_t::is_foreground() const { return group->is_foreground(); }
 
 maybe_t<pid_t> job_t::get_pgid() const { return group->get_pgid(); }
 
+maybe_t<pid_t> job_t::get_last_pid() const {
+    for (auto iter = processes.rbegin(); iter != processes.rend(); ++iter) {
+        const process_t *proc = iter->get();
+        if (proc->pid > 0) return proc->pid;
+    }
+    return none();
+}
+
 job_id_t job_t::job_id() const { return group->get_id(); }
 
 void job_t::continue_job(parser_t &parser, bool in_foreground) {
     // Put job first in the job list.
     parser.job_promote(this);
-    mut_flags().notified = false;
+    mut_flags().notified_of_stop = false;
 
     int pgid = -2;
     if (auto tmp = get_pgid()) pgid = *tmp;
