@@ -8,7 +8,7 @@
 #include <signal.h>
 #include <stddef.h>
 #include <sys/time.h>  // IWYU pragma: keep
-#include <sys/wait.h>
+#include <sys/wait.h>  // IWYU pragma: keep
 #include <unistd.h>
 
 #include <deque>
@@ -21,9 +21,10 @@
 #include "io.h"
 #include "parse_tree.h"
 #include "topic_monitor.h"
+#include "wait_handle.h"
 
 /// Types of processes.
-enum class process_type_t {
+enum class process_type_t : uint8_t {
     /// A regular external command.
     external,
     /// A builtin command.
@@ -36,7 +37,7 @@ enum class process_type_t {
     exec,
 };
 
-enum class job_control_t {
+enum class job_control_t : uint8_t {
     all,
     interactive,
     none,
@@ -197,16 +198,7 @@ enum { INVALID_PID = -2 };
 /// If the process is of type process_type_t::function, argv is the argument vector, and argv[0] is
 /// the name of the shellscript function.
 class parser_t;
-class process_t {
-   private:
-    null_terminated_array_t<wchar_t> argv_array;
-
-    redirection_spec_list_t proc_redirection_specs;
-
-    // No copying.
-    process_t(const process_t &rhs) = delete;
-    void operator=(const process_t &rhs) = delete;
-
+class process_t : noncopyable_t {
    public:
     process_t();
 
@@ -230,30 +222,19 @@ class process_t {
     std::vector<concrete_assignment> variable_assignments;
 
     /// Sets argv.
-    void set_argv(const wcstring_list_t &argv) { argv_array.set(argv); }
+    void set_argv(wcstring_list_t argv) { argv_ = std::move(argv); }
 
     /// Returns argv.
-    wchar_t **get_argv() { return argv_array.get(); }
-    const wchar_t *const *get_argv() const { return argv_array.get(); }
-    const null_terminated_array_t<wchar_t> &get_argv_array() const { return argv_array; }
+    const wcstring_list_t &argv() { return argv_; }
 
-    /// Returns argv[idx].
-    const wchar_t *argv(size_t idx) const {
-        const wchar_t *const *argv = argv_array.get();
-        assert(argv != nullptr);
-        return argv[idx];
-    }
-
-    /// Returns argv[0], or NULL.
-    const wchar_t *argv0() const {
-        const wchar_t *const *argv = argv_array.get();
-        return argv ? argv[0] : nullptr;
-    }
+    /// Returns argv[0], or nullptr.
+    const wchar_t *argv0() const { return argv_.empty() ? nullptr : argv_.front().c_str(); }
 
     /// Redirection list getter and setter.
-    const redirection_spec_list_t &redirection_specs() const { return proc_redirection_specs; }
+    const redirection_spec_list_t &redirection_specs() const { return proc_redirection_specs_; }
+
     void set_redirection_specs(redirection_spec_list_t specs) {
-        this->proc_redirection_specs = std::move(specs);
+        this->proc_redirection_specs_ = std::move(specs);
     }
 
     /// Store the current topic generations. That is, right before the process is launched, record
@@ -268,6 +249,14 @@ class process_t {
     /// \return whether this process type is internal (block, function, or builtin).
     bool is_internal() const;
 
+    /// \return the wait handle for the process, if it exists.
+    wait_handle_ref_t get_wait_handle() { return wait_handle_; }
+
+    /// Create a wait handle for the process.
+    /// As a process does not know its job id, we pass it in.
+    /// Note this will return null if the process is not waitable (has no pid).
+    wait_handle_ref_t make_wait_handle(internal_job_id_t jid);
+
     /// Actual command to pass to exec in case of process_type_t::external or process_type_t::exec.
     wcstring actual_cmd;
 
@@ -292,25 +281,25 @@ class process_t {
     /// Reported status value.
     proc_status_t status{};
 
-    /// Last time of cpu time check.
-    struct timeval last_time {};
+    /// Last time of cpu time check, in seconds (per timef).
+    timepoint_t last_time{0};
 
     /// Number of jiffies spent in process at last cpu time check.
     unsigned long last_jiffies{0};
+
+   private:
+    wcstring_list_t argv_;
+    redirection_spec_list_t proc_redirection_specs_;
+
+    // The wait handle. This is constructed lazily, and cached.
+    wait_handle_ref_t wait_handle_{};
 };
 
-typedef std::unique_ptr<process_t> process_ptr_t;
-typedef std::vector<process_ptr_t> process_list_t;
-
-/// A user-visible job ID.
-using job_id_t = int;
-
-/// The non user-visible, never-recycled job ID.
-/// Every job has a unique positive value for this.
-using internal_job_id_t = uint64_t;
+using process_ptr_t = std::unique_ptr<process_t>;
+using process_list_t = std::vector<process_ptr_t>;
 
 /// A struct representing a job. A job is a pipeline of one or more processes.
-class job_t {
+class job_t : noncopyable_t {
    public:
     /// A set of jobs properties. These are immutable: they do not change for the lifetime of the
     /// job.
@@ -342,10 +331,6 @@ class job_t {
     /// messages about job status on the terminal.
     const wcstring command_str;
 
-    // No copying.
-    job_t(const job_t &rhs) = delete;
-    void operator=(const job_t &) = delete;
-
    public:
     job_t(const properties_t &props, wcstring command_str);
     ~job_t();
@@ -396,6 +381,11 @@ class job_t {
     /// This may also be fish itself.
     maybe_t<pid_t> get_pgid() const;
 
+    /// \return the pid of the last external process in the job.
+    /// This may be none if the job consists of just internal fish functions or builtins.
+    /// This will never be fish's own pid.
+    maybe_t<pid_t> get_last_pid() const;
+
     /// The id of this job.
     /// This is user-visible, is recycled, and may be -1.
     job_id_t job_id() const;
@@ -409,10 +399,11 @@ class job_t {
         /// forked, etc.
         bool constructed{false};
 
-        /// Whether the user has been told about stopped job.
-        bool notified{false};
+        /// Whether the user has been notified that this job is stopped (if it is).
+        bool notified_of_stop{false};
 
         /// Whether the exit status should be negated. This flag can only be set by the not builtin.
+        /// Two "not" prefixes on a single job cancel each other out.
         bool negate{false};
 
         /// This job is disowned, and should be removed from the active jobs list.
@@ -465,9 +456,8 @@ class job_t {
     /// \return whether this job's group is in the foreground.
     bool is_foreground() const;
 
-    /// \return whether we should report process exit events.
-    /// This implements some historical behavior which has not been justified.
-    bool should_report_process_exits() const;
+    /// \return whether we should post job_exit events.
+    bool posts_job_exit_events() const;
 
     /// \return whether this job and its parent chain are fully constructed.
     bool job_chain_is_fully_constructed() const;
@@ -485,6 +475,7 @@ class job_t {
     /// \returns the statuses for this job.
     maybe_t<statuses_t> get_statuses() const;
 };
+using job_ref_t = std::shared_ptr<job_t>;
 
 /// Whether this shell is attached to a tty.
 bool is_interactive_session();
@@ -501,7 +492,7 @@ bool no_exec();
 void mark_no_exec();
 
 // List of jobs.
-typedef std::deque<shared_ptr<job_t>> job_list_t;
+using job_list_t = std::deque<job_ref_t>;
 
 /// The current job control mode.
 ///
@@ -523,9 +514,9 @@ job_list_t jobs_requiring_warning_on_exit(const parser_t &parser);
 /// jobs_requiring_warning_on_exit().
 void print_exit_warning_for_jobs(const job_list_t &jobs);
 
-/// Use the procfs filesystem to look up how many jiffies of cpu time was used by this process. This
+/// Use the procfs filesystem to look up how many jiffies of cpu time was used by a given pid. This
 /// function is only available on systems with the procfs file entry 'stat', i.e. Linux.
-unsigned long proc_get_jiffies(process_t *p);
+unsigned long proc_get_jiffies(pid_t inpid);
 
 /// Update process time usage for all processes by calling the proc_get_jiffies function for every
 /// process of every job.
@@ -535,9 +526,6 @@ void proc_update_jiffies(parser_t &parser);
 /// job is in the foreground, that every process is in a valid state, etc.
 void proc_sanity_check(const parser_t &parser);
 
-/// Create a process/job exit event notification.
-event_t proc_create_event(const wchar_t *msg, event_type_t type, pid_t pid, int status);
-
 /// Initializations.
 void proc_init();
 
@@ -564,7 +552,7 @@ int terminal_maybe_give_to_job_group(const job_group_t *jg, bool continuing_from
 
 /// Add a job to the list of PIDs/PGIDs we wait on even though they are not associated with any
 /// jobs. Used to avoid zombie processes after disown.
-void add_disowned_job(job_t *j);
+void add_disowned_job(const job_t *j);
 
 bool have_proc_stat();
 
