@@ -2,13 +2,14 @@
 #include "config.h"  // IWYU pragma: keep
 
 #include <errno.h>
-#include <signal.h>
 #include <stdio.h>
 #ifdef HAVE_SIGINFO_H
 #include <siginfo.h>
 #endif
 #include <pthread.h>
 
+#include <csignal>
+
 #include "common.h"
 #include "event.h"
 #include "fallback.h"  // IWYU pragma: keep
@@ -270,7 +271,6 @@ static void fish_signal_handler(int sig, siginfo_t *info, void *context) {
             // We have a sigalarm handler that does nothing. This is used in the signal torture
             // test, to verify that we behave correctly when receiving lots of irrelevant signals.
             break;
-
     }
     errno = saved_errno;
 }
